"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import math
from collections import deque

import torch
from functools import partial

import networkx as nx
from nncf.layers import NNCF_CONV_MODULES_DICT

from nncf.dynamic_graph.context import Scope
from nncf.dynamic_graph.graph import NNCFGraph, NNCFNode
from nncf.nncf_network import NNCFNetwork


# pylint: disable=protected-access
def get_rounded_pruned_element_number(total, sparsity_rate, multiple_of=8):
    """
    Calculates number of sparsified elements (approximately sparsity rate) from total such as
    number of remaining items will be multiple of some value.
    Always rounds number of remaining elements up.
    :param total: total elements number
    :param sparsity_rate: prorortion of zero elements in total.
    :param multiple_of:
    :return: number of elements to be zeroed
    """
    remaining_elems = math.ceil((total - total * sparsity_rate) / multiple_of) * multiple_of
    return max(total - remaining_elems, 0)


def get_bn_node_for_conv(graph: nx.Graph, conv_node: dict):
    out_edges = graph.out_edges(conv_node['key'])
    for _, out_node_key in out_edges:
        out_node = graph.nodes[out_node_key]
        if out_node['op_exec_context'].operator_name == 'batch_norm':
            return out_node
    return None


def get_bn_for_module_scope(target_model: NNCFNetwork, module_scope: Scope):
    """
    Returns batch norm module that corresponds to module_scope convolution.
    :param target_model: NNCFNetwork to work with
    :param module_scope:
    :return: batch norm module
    """
    graph = target_model.get_original_graph()
    module_graph_node = graph.find_node_in_nx_graph_by_scope(module_scope)
    bn_graph_node = get_bn_node_for_conv(graph._nx_graph, module_graph_node)
    bn_module = None
    if bn_graph_node:
        bn_module = target_model.get_module_by_scope(bn_graph_node['op_exec_context'].scope_in_model)
    return bn_module


def find_first_ops_with_type(nncf_graph: NNCFGraph, nodes, required_types, forward: bool = True):
    """
    Looking for first nodes with type from pruned_ops_types that are reachable from nodes.
    :param nncf_graph: NNCFGraph to work with
    :param nodes: nodes from which search begins
    :param required_types: types of nodes for search
    :param forward: whether the search will be forward or backward
    :return:
    """
    graph = nncf_graph._nx_graph
    get_edges_fn = graph.out_edges if forward else graph.in_edges

    found_nodes = []
    visited = {n: False for n in graph.nodes}
    node_stack = deque(nodes)
    while node_stack:
        last_node = node_stack.pop()
        last_node_type = nncf_graph.node_type_fn(last_node)

        if not visited[last_node['key']]:
            visited[last_node['key']] = True
        else:
            continue

        if last_node_type not in required_types:
            edges = get_edges_fn(last_node['key'])
            for in_node_name, out_node_name in edges:
                cur_node = graph.nodes[out_node_name] if forward else graph.nodes[in_node_name]

                if not visited[cur_node['key']]:
                    node_stack.append(cur_node)
        else:
            found_nodes.append(last_node)
    return found_nodes


def traverse_function(node: NNCFNode, output, nncf_graph: NNCFGraph, required_types, visited):
    nx_node = nncf_graph._nx_graph.nodes[nncf_graph.get_node_key_by_id(node.node_id)]
    node_type = nncf_graph.node_type_fn(nx_node)
    if visited[node.node_id]:
        return True, output
    visited[node.node_id] = True

    if node_type not in required_types:
        return False, output

    output.append(node)
    return True, output


def get_first_pruned_modules(target_model: NNCFNetwork, pruned_ops_types):
    """
    Looking for first pruned modules in target model.
    First == layer of pruned type, that there is a path from the input such that there are no other
    pruned operations on it.
    :param pruned_ops_types: types of modules that will be pruned
    :param target_model: model to work with
    :return: list of all first pruned modules
    """
    graph = target_model.get_original_graph()  # NNCFGraph here
    graph_roots = graph.get_input_nodes()  # NNCFNodes here

    visited = {node_id: False for node_id in graph.get_all_node_idxs()}
    partial_traverse_function = partial(traverse_function, nncf_graph=graph, required_types=pruned_ops_types,
                                        visited=visited)

    first_pruned_nodes = []
    for root in graph_roots:
        first_pruned_nodes.extend(graph.traverse_graph(root, partial_traverse_function))
    first_pruned_modules = [target_model.get_module_by_scope(n.op_exec_context.scope_in_model)
                            for n in first_pruned_nodes]
    return first_pruned_modules


def get_last_pruned_modules(target_model: NNCFNetwork, pruned_ops_types):
    """
    Looking for last pruned modules in target model.
    Last == layer of pruned type, that there is a path from this layer to the model output
    such that there are no other pruned operations on it.
    :param pruned_ops_types: types of modules that will be pruned
    :param target_model: model to work with
    :return: list of all last pruned modules
    """
    graph = target_model.get_original_graph()  # NNCFGraph here
    graph_outputs = graph.get_graph_outputs()  # NNCFNodes here

    visited = {node_id: False for node_id in graph.get_all_node_idxs()}
    partial_traverse_function = partial(traverse_function, nncf_graph=graph, required_types=pruned_ops_types,
                                        visited=visited)
    last_pruned_nodes = []
    for output in graph_outputs:
        last_pruned_nodes.extend(graph.traverse_graph(output, partial_traverse_function, False))

    last_pruned_modules = [target_model.get_module_by_scope(n.op_exec_context.scope_in_model)
                           for n in last_pruned_nodes]
    return last_pruned_modules


def get_sources_of_node(nncf_node: NNCFNode, graph: NNCFGraph, sources_types):
    """
    Source is a node of sourse such that there is path from this node to nx_node and on this path
    no node has one of sources_types type.
    :param sources_types: list of sources types
    :param nncf_node: NNCFNode to get sources
    :param graph: NNCF graph to work with
    :return: list of all sources nodes
    """
    visited = {node_id: False for node_id in graph.get_all_node_idxs()}
    partial_traverse_function = partial(traverse_function, nncf_graph=graph, required_types=sources_types,
                                        visited=visited)
    nncf_nodes = [nncf_node]
    if nncf_node.op_exec_context.operator_name in sources_types:
        nncf_nodes = graph.get_previous_nodes(nncf_node)

    source_nodes = []
    for node in nncf_nodes:
        source_nodes.extend(graph.traverse_graph(node, partial_traverse_function, False))
    return source_nodes


def is_conv_with_downsampling(conv_module):
    return not torch.all(torch.tensor(conv_module.stride) == 1)


def is_grouped_conv(conv_module):
    return conv_module.groups != 1


def is_depthwise_conv(conv_module):
    return conv_module.groups == conv_module.in_channels and (conv_module.out_channels % conv_module.in_channels == 0)


def get_previous_conv(target_model: NNCFNetwork, module, module_scope):
    """
    Return source convolution of module. If node has other source type or there are more than one source - return None.
    """
    conv_types = [str.lower(v.__name__) for v in NNCF_CONV_MODULES_DICT.values()]

    graph = target_model.get_original_graph()
    nx_node = graph.find_node_in_nx_graph_by_scope(module_scope)
    nncf_node = graph._nx_node_to_nncf_node(nx_node)
    sources = get_sources_of_node(nncf_node, graph, conv_types + ['linear'])
    if len(sources) == 1 and sources[0].op_exec_context.operator_name in conv_types:
        return sources[0]
    return None



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from typing import List

from torch import Tensor


class TracesOrder:
    def __init__(self, execution_indexes_of_weights_ordered_by_traces: List[int]):
        self._index_by_traces_to_execution_index = execution_indexes_of_weights_ordered_by_traces
        self._num_weights = len(execution_indexes_of_weights_ordered_by_traces)
        self._index_by_execution_to_index_by_traces = \
            [execution_indexes_of_weights_ordered_by_traces.index(i) for i in range(self._num_weights)]

    def get_execution_order_config(self, bitwidth_by_traces: List[int]) -> List[int]:
        execution_order_config = [0] * len(bitwidth_by_traces)
        for i, bitwidth in enumerate(bitwidth_by_traces):
            execution_order_config[self._index_by_traces_to_execution_index[i]] = bitwidth
        return execution_order_config

    def get_traces_order_config(self, bitwidth_by_execution: List[int]) -> List[int]:
        traces_order_config = [0] * len(bitwidth_by_execution)
        for i, bitwidth in enumerate(bitwidth_by_execution):
            traces_order_config[self._index_by_execution_to_index_by_traces[i]] = bitwidth
        return traces_order_config

    def get_execution_index_by_traces_index(self, traces_index: int):
        return self._index_by_traces_to_execution_index[traces_index]

    def __bool__(self):
        return bool(self._index_by_traces_to_execution_index)


class TracesPerLayer:
    def __init__(self, traces_per_layer_by_execution: Tensor):
        self._traces_per_layer_by_execution = traces_per_layer_by_execution
        execution_indexes_of_weights_in_descending_order_of_traces = \
            [i[0] for i in sorted(enumerate(traces_per_layer_by_execution), reverse=False, key=lambda x: x[1])]
        self.traces_order = TracesOrder(execution_indexes_of_weights_in_descending_order_of_traces)

    def get_by_execution_index(self, execution_index: int) -> Tensor:
        return self._traces_per_layer_by_execution[execution_index]

    def get_by_trace_index(self, trace_index: int) -> Tensor:
        execution_index = self.traces_order.get_execution_index_by_traces_index(trace_index)
        return self._traces_per_layer_by_execution[execution_index]

    def get_all(self) -> Tensor:
        return self._traces_per_layer_by_execution

    def __bool__(self):
        return bool(self.traces_order)



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import numpy as np
import scipy.optimize
from nncf.compression_method_api import CompressionScheduler
from nncf.config import NNCFConfig
from nncf.registry import Registry

PRUNING_SCHEDULERS = Registry("pruning_schedulers")


class PruningScheduler(CompressionScheduler):
    def __init__(self, pruning_algo, params: NNCFConfig = None):
        super().__init__()
        if params is None:
            self._params = NNCFConfig()
        else:
            self._params = params

        self.algo = pruning_algo

        # Number of initial steps of training before pruning
        self.num_init_steps = self._params.get('num_init_steps', 0)
        self.pruning_steps = self._params.get('pruning_steps', 100)

        # Pruning rates
        self.initial_pruning = self._params.get('pruning_init', 0)
        self.pruning_target = self._params.get('pruning_target', 0.5)

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self._set_pruning_level()

    def epoch_step(self, epoch=None):
        super().epoch_step(epoch)
        self._set_pruning_level()

    def _set_pruning_level(self):
        self.algo.set_pruning_rate(self.current_pruning_level)

        curr_epoch = self.last_epoch + 1
        if curr_epoch >= (self.pruning_steps + self.num_init_steps):
            self.algo.freeze()

    def _calc_pruning_level(self):
        raise NotImplementedError

    @property
    def current_pruning_level(self):
        curr_epoch = self.last_epoch + 1
        if curr_epoch >= self.num_init_steps:
            return self._calc_pruning_level()
        return 0

    def _calc_density_level(self):
        return 1 - self.current_pruning_level()

@PRUNING_SCHEDULERS.register("baseline")
class BaselinePruningScheduler(PruningScheduler):
    """
    Baseline scheduler that setting max pruning rate after num_init_steps epoch
    and freeze algorithm after it.
    """
    def __init__(self, pruning_algo, config=None):
        super().__init__(pruning_algo, config)
        self._set_pruning_level()

    def _calc_pruning_level(self):
        return self.pruning_target

    def _set_pruning_level(self):
        self.algo.set_pruning_rate(self.current_pruning_level)
        curr_epoch = self.last_epoch + 1
        if curr_epoch >= self.num_init_steps:
            self.algo.freeze()


@PRUNING_SCHEDULERS.register("exponential")
class ExponentialPruningScheduler(PruningScheduler):
    """
    Calculates pruning rate progressively according to the formula
    P = 1 - a * exp(- k * epoch)
    Where:
    epoch - epoch number
    P - pruning rate for current epoch
    a, k - params
    """
    def __init__(self, pruning_algo, config=None):
        super().__init__(pruning_algo, config)
        self.a, self.k = self._init_exp(self.initial_pruning, self.pruning_target, pruning_steps=self.pruning_steps)
        self._set_pruning_level()

    def _calc_pruning_level(self):
        curr_epoch = self.last_epoch + 1
        curr_pruning = 1 - self.a * np.exp(-self.k * (curr_epoch - self.num_init_steps))
        max_pruning = self.pruning_target
        return max_pruning if curr_pruning >= max_pruning else curr_pruning

    @staticmethod
    def _init_exp(initial_pruning, max_pruning, pruning_steps=20):
        p1 = (0, 1 - initial_pruning)
        p2 = (pruning_steps, 1 - max_pruning)
        k = np.log(p2[1] / p1[1]) / (p1[0] - p2[0])
        a = p1[1] / np.exp(-k * p1[0])
        return a, k


@PRUNING_SCHEDULERS.register("exponential_with_bias")
class ExponentialWithBiasPruningScheduler(PruningScheduler):
    """
    Calculates pruning rate progressively according to the formula
    P = a * exp(- k * epoch) + b
    Where:
    epoch - epoch number
    P - pruning rate for current epoch
    a, b, k - params
    """
    def __init__(self, pruning_algo, config=None):
        super().__init__(pruning_algo, config)
        self.a, self.b, self.k = self._init_exp(self.pruning_steps, self.initial_pruning, self.pruning_target)
        self._set_pruning_level()

    def _calc_pruning_level(self):
        curr_epoch = self.last_epoch
        curr_pruning = self.a * np.exp(-self.k * (curr_epoch - self.num_init_steps)) + self.b
        max_pruning = self.pruning_target
        return max_pruning if curr_pruning >= max_pruning else curr_pruning

    @staticmethod
    def _init_exp(E_max, P_min, P_max, D=1 / 8):
        """
        Find a, b, k for system (from SPFP paper):
        1. P_min = a + b
        2. P_max = a * exp(-k * E_max) + b
        3. 3/4 * P_max = a *  exp(-k * E_max * D) + b
        Where P_min, P_max - minimal and goal levels of pruning rate
        E_max - number of epochs for pruning
        """
        def get_b(a, k):
            return P_min - a

        def get_a(k):
            return (3 / 4 * P_max - P_min) / (np.exp(- D * k * E_max) - 1)

        def f_to_solve(x):
            y = np.exp(D * x * E_max)
            return 1 / 3 * y + 1 / (y ** 7) - 4 / 3

        k = scipy.optimize.fsolve(f_to_solve, [1])[0]
        a = get_a(k)
        b = get_b(a, k)
        return a, b, k



"""
 Copyright (c) 2019-2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import numpy as np
from bisect import bisect_right

from nncf.nncf_logger import logger
from ..algo_selector import Registry
from ..compression_method_api import CompressionScheduler, CompressionLevel

SPARSITY_SCHEDULERS = Registry("sparsity_schedulers")


class SparsityScheduler(CompressionScheduler):
    def __init__(self, sparsity_algo, params: dict = None):
        super().__init__()
        if params is None:
            self._params = dict()
        else:
            self._params = params

        self.algo = sparsity_algo
        self.initial_sparsity = self._params.get('sparsity_init', 0)
        self.sparsity_target = self._params.get('sparsity_target', 0.5)
        self.sparsity_target_epoch = self._params.get('sparsity_target_epoch', 90)
        self.sparsity_freeze_epoch = self._params.get('sparsity_freeze_epoch', 100)

    def initialize(self):
        self._set_sparsity_level()

    def epoch_step(self, epoch=None):
        super().epoch_step(epoch)
        self._set_sparsity_level()

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self._set_sparsity_level()

    def _set_sparsity_level(self):
        self.algo.set_sparsity_level(self.current_sparsity_level)
        if self.last_epoch + 1 >= self.sparsity_freeze_epoch:
            self.algo.freeze()

    def _calc_density_level(self):
        return 1 - self.current_sparsity_level

    @property
    def current_sparsity_level(self):
        raise NotImplementedError

    @property
    def target_sparsity_level(self) -> float:
        return self.sparsity_target

    def compression_level(self) -> CompressionLevel:
        if self.current_sparsity_level == 0:
            return CompressionLevel.NONE
        if self.current_sparsity_level >= self.target_sparsity_level:
            return CompressionLevel.FULL
        return CompressionLevel.PARTIAL


@SPARSITY_SCHEDULERS.register("polynomial")
class PolynomialSparseScheduler(SparsityScheduler):
    def __init__(self, sparsity_algo, params=None):
        super().__init__(sparsity_algo, params)
        self.power = self._params.get('power', 0.9)
        self.concave = self._params.get('concave', False)
        self._update_per_optimizer_step = self._params.get('update_per_optimizer_step', False)
        if self._update_per_optimizer_step:
            self._steps_per_epoch = self._params.get('steps_per_epoch')
            if self._steps_per_epoch is None:
                logger.warning("Optimizer set to update sparsity level per optimizer step,"
                               "but steps_per_epoch was not set in config. Will only start updating "
                               "sparsity level after measuring the actual steps per epoch as signaled "
                               "by a .epoch_step() call.")

        self._set_sparsity_level()

    def step(self, last=None):
        super().step(last)
        if self._update_per_optimizer_step and self._steps_per_epoch is not None:
            self._set_sparsity_level()

    def epoch_step(self, epoch=None):
        if self._update_per_optimizer_step:
            if self.last_epoch == -1 and self._steps_per_epoch is None:
                self._steps_per_epoch = self._steps_in_current_epoch

                # Reset step and epoch step counters
                self.last_step = -1
                epoch = -1
            if self._steps_in_current_epoch != self._steps_per_epoch:
                self._steps_per_epoch = self._steps_in_current_epoch
                logger.warning("Actual optimizer steps per epoch is different than what is "
                               "specified by scheduler parameters! Scheduling may be incorrect. "
                               "Setting scheduler's global step count to (current epoch) * "
                               "(actual steps per epoch)")
                self.step(self._steps_per_epoch * self.last_epoch)

        self._steps_in_current_epoch = 0
        super().epoch_step(epoch)

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self._set_sparsity_level()

    @property
    def current_sparsity_level(self):
        epochs_completed = self.last_epoch + 1
        if self._update_per_optimizer_step:
            if self._steps_per_epoch is None:
                return self.initial_sparsity  # Cannot do proper sparsity update until the steps in an epoch are counted
            fractional_epoch = epochs_completed + self._steps_in_current_epoch / self._steps_per_epoch
            progress = (min(self.sparsity_target_epoch, fractional_epoch) / self.sparsity_target_epoch)
        else:
            progress = (min(self.sparsity_target_epoch, epochs_completed) / self.sparsity_target_epoch)

        if self.concave:
            current_sparsity = self.initial_sparsity + (self.sparsity_target - self.initial_sparsity) * (
                progress ** self.power)
        else:
            current_sparsity = self.sparsity_target - (self.sparsity_target - self.initial_sparsity) * (
                (1 - progress) ** self.power)
        return current_sparsity


@SPARSITY_SCHEDULERS.register("exponential")
class ExponentialSparsityScheduler(SparsityScheduler):
    def __init__(self, sparsity_algo, params=None):
        super().__init__(sparsity_algo, params)
        self.a, self.k = self._init_exp(self.initial_sparsity, self.sparsity_target,
                                        sparsity_steps=self.sparsity_target_epoch)
        self._set_sparsity_level()

    @property
    def current_sparsity_level(self):
        curr_epoch = self.last_epoch + 1
        curr_sparsity = 1 - self.a * np.exp(-self.k * curr_epoch)
        return curr_sparsity if curr_sparsity <= self.sparsity_target else self.sparsity_target

    @staticmethod
    def _init_exp(initial_sparsity, max_sparsity, sparsity_steps=20):
        p1 = (0, 1 - initial_sparsity)
        p2 = (sparsity_steps, 1 - max_sparsity)
        k = np.log(p2[1] / p1[1]) / (p1[0] - p2[0])
        a = p1[1] / np.exp(-k * p1[0])
        return a, k


@SPARSITY_SCHEDULERS.register("adaptive")
class AdaptiveSparsityScheduler(SparsityScheduler):
    def __init__(self, sparsity_algo, params=None):
        super().__init__(sparsity_algo, params)
        self.sparsity_loss = sparsity_algo.loss
        from .rb.loss import SparseLoss
        if not isinstance(self.sparsity_loss, SparseLoss):
            raise TypeError('AdaptiveSparseScheduler expects SparseLoss, but {} is given'.format(
                self.sparsity_loss.__class__.__name__))
        self.decay_step = params.get('step', 0.05)
        self.eps = params.get('eps', 0.03)
        self.patience = params.get('patience', 1)
        self.current_sparsity_target = self.initial_sparsity
        self.num_bad_epochs = 0
        self._set_sparsity_level()

    def epoch_step(self, epoch=None):
        super().epoch_step(epoch)
        if self.sparsity_loss.current_sparsity >= self.current_sparsity_target - self.eps:
            self.num_bad_epochs += 1

        if self.num_bad_epochs >= self.patience:
            self.num_bad_epochs = 0
            self.current_sparsity_target = min(self.current_sparsity_target + self.decay_step, self.sparsity_target)
        self._set_sparsity_level()

    def state_dict(self):
        sd = super().state_dict()
        sd['num_bad_epochs'] = self.num_bad_epochs
        sd['current_sparsity_level'] = self.current_sparsity_level
        return sd

    @property
    def current_sparsity_level(self):
        return self.current_sparsity_target


@SPARSITY_SCHEDULERS.register("multistep")
class MultiStepSparsityScheduler(SparsityScheduler):
    def __init__(self, sparsity_algo, params):
        super().__init__(sparsity_algo, params)
        self.sparsity_levels = self._params.get('multistep_sparsity_levels', [0.1, 0.5])
        self.steps = self._params.get('multistep_steps', [90])
        if len(self.steps) + 1 != len(self.sparsity_levels):
            raise AttributeError('number of sparsity levels must equal to number of steps + 1')

        self.initial_sparsity = self.sparsity_level = self.sparsity_levels[0]
        self.max_sparsity = max(self.sparsity_levels)
        self.sparsity_algo = sparsity_algo
        self.steps = sorted(self.steps)
        self.max_step = self.steps[-1]
        self.prev_ind = 0
        self._set_sparsity_level()

    def epoch_step(self, last=None):
        super().epoch_step(last)
        curr_epoch = self.last_epoch + 1
        ind = bisect_right(self.steps, curr_epoch)
        if ind != self.prev_ind:
            self.sparsity_level = self.sparsity_levels[ind]
            self.prev_ind = ind
        self._set_sparsity_level()

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        ind = bisect_right(self.steps, self.last_epoch)
        if ind > 0:
            self.prev_ind = ind
            self.sparsity_level = self.sparsity_levels[ind]
            self._set_sparsity_level()

    @property
    def current_sparsity_level(self):
        return self.sparsity_level

    @property
    def target_sparsity_level(self):
        return self.sparsity_levels[-1]



"""
 Copyright (c) 2019-2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import inspect
from collections import OrderedDict, Counter
from enum import Enum
from typing import List, Callable, Tuple, Dict, Optional

import functools
import networkx as nx
import numpy as np
import torch
from copy import deepcopy
from torch import nn

from nncf.debug import CombinedDebugInterface, debuggable_forward, is_debug
from nncf.dynamic_graph.context import TracingContext
from nncf.dynamic_graph.graph import NNCFGraph, InputAgnosticOperationExecutionContext, OperationExecutionContext
from nncf.dynamic_graph.graph import ShapeIgnoringTensorMetaComparator
from nncf.dynamic_graph.graph_builder import GraphBuilder, PostGraphBuildActing, create_dummy_forward_fn, \
    ModelInputInfo
from nncf.dynamic_graph.graph_matching import NodeExpression
from nncf.dynamic_graph.operator_metatypes import OPERATOR_METATYPES
from nncf.dynamic_graph.patch_pytorch import ignore_scope
from nncf.dynamic_graph.input_wrapping import InputInfoWrapManager, MODEL_INPUT_OP_NAME
from nncf.dynamic_graph.transform_graph import replace_modules_by_nncf_modules
from nncf.hw_config import HWConfig
from nncf.layers import NNCF_MODULES, NNCF_WRAPPED_USER_MODULES_DICT
from nncf.nncf_logger import logger as nncf_logger
from nncf.quantization.layers import QUANTIZATION_MODULES
from nncf.utils import get_all_modules_by_type, get_state_dict_names_with_modules

MODEL_WRAPPED_BY_NNCF_ATTR_NAME = 'nncf_module'


class CompressionModuleType(Enum):
    FUNCTION_QUANTIZER = 0
    ACTIVATION_QUANTIZER = 1


@functools.total_ordering
class OperationPriority(Enum):
    DEFAULT_PRIORITY = 0
    SPARSIFICATION_PRIORITY = 2
    QUANTIZATION_PRIORITY = 11
    PRUNING_PRIORITY = 1

    def __lt__(self, other):
        # pylint: disable=comparison-with-callable
        return self.value < other.value


class InsertionType(Enum):
    OPERATOR_PRE_HOOK = 0
    OPERATOR_POST_HOOK = 1
    NNCF_MODULE_PRE_OP = 2
    NNCF_MODULE_POST_OP = 3

    def __eq__(self, other):
        # pylint: disable=comparison-with-callable
        if isinstance(other, InsertionType):
            return self.value == other.value
        return self.value == other


class InsertionInfo:
    def __init__(self, op_exec_context: OperationExecutionContext,
                 is_input=False,
                 is_output=False,
                 shape_to_operate_on=None):
        self.op_exec_context = op_exec_context  # type: OperationExecutionContext
        self.is_input = is_input
        self.is_output = is_output
        self.shape_to_operate_on = shape_to_operate_on
        self.linked_op_exec_contexts = []  # type: List[OperationExecutionContext]

    def __eq__(self, other: 'InsertionInfo'):
        return self.op_exec_context == other.op_exec_context and Counter(self.linked_op_exec_contexts) == Counter(
            other.linked_op_exec_contexts)

    def __hash__(self):
        return self.op_exec_context.__hash__()


class InsertionPoint:
    def __init__(self, ia_op_exec_context: InputAgnosticOperationExecutionContext,
                 insertion_type: InsertionType):
        self.ia_op_exec_context = ia_op_exec_context
        self.insertion_type = insertion_type

    def __eq__(self, other: 'InsertionPoint'):
        return self.insertion_type == other.insertion_type and self.ia_op_exec_context == other.ia_op_exec_context

    def __str__(self):
        return str(self.insertion_type) + " " + str(self.ia_op_exec_context)

    def __hash__(self):
        return hash(str(self))


class InsertionCommand:
    def __init__(self, point: InsertionPoint, fn: Callable,
                 priority: OperationPriority = OperationPriority.DEFAULT_PRIORITY):
        self.insertion_point = point  # type: InsertionPoint
        self.fn = fn  # type: Callable
        self.priority = priority  # type: OperationPriority


class LoadStateListener:
    """
        Resets the initialization flags (`initialized`) for all quantization modules on `load_state_dict` call.
        These flags are used to update not loaded params (from checkpoint or model's state)
        on initialization stage of algorithm.
        Flags reset is required on each call of `load_state_dict`, because internal method (`build_graph`)
        restores model state by calling this method.
    """

    def __init__(self, model, all_quantizations):
        for prefix, module in all_quantizations.items():
            module.state_dict_name = prefix
        # pylint: disable=protected-access
        self.hook = model._register_load_state_dict_pre_hook(
            functools.partial(self.hook_fn, quantize_modules=all_quantizations.values()))

    def hook_fn(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs,
                quantize_modules):
        for module in quantize_modules:
            module.initialized = False

    def close(self):
        self.hook.remove()


class InsertionPointGraphNodeType(Enum):
    INSERTION_POINT = 0
    OPERATOR = 1


class InsertionPointGraph(nx.DiGraph):
    """
    This graph is built from the NNCFGraph representation of the model control flow graph and adds ephemeral
    "insertion point nodes" into the NNCF model graph representation corresponding to operator pre- and
    post-hooks. Module pre-op and post-op insertion points are currently not reflected here, but they are
    probably not required for quantizing activations, for which the quantizer propagation makes sense.
    This "insertion point graph" representation is useful for quantizer propagation and for referencing
    the compression algorithm hooks to the model operations to which they are applied to.
    """
    NODE_TYPE_NODE_ATTR = "node_type"
    INSERTION_POINT_DATA_NODE_ATTR = "insertion_point_data"
    IS_IN_NNCF_MODULE_NODE_ATTR = "is_in_nncf_module"
    REGULAR_NODE_REF_NODE_ATTR = "regular_node_ref"
    ASSOCIATED_IP_NODE_KEYS_NODE_ATTR = "associated_ip_node_keys"
    OPERATOR_METATYPE_NODE_ATTR = "op_meta"

    PRE_HOOK_ID_PREFIX = "PRE HOOK "  # NB: Do not use colon (':') in node keys! Causes trouble for .dot file export.
    POST_HOOK_ID_PREFIX = "POST HOOK "

    def __init__(self, model_nx_graph: nx.DiGraph):
        super().__init__()
        self._base_nx_graph = deepcopy(model_nx_graph)

        for node_key, node in self._base_nx_graph.nodes.items():
            attrs = {InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR: node,
                     InsertionPointGraph.NODE_TYPE_NODE_ATTR: InsertionPointGraphNodeType.OPERATOR,
                     InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR: set(),
                     InsertionPointGraph.OPERATOR_METATYPE_NODE_ATTR: None}
            self.add_node(node_key, **attrs)
        for from_node, to_node in self._base_nx_graph.edges:
            self.add_edge(from_node, to_node)

        # TODO: Add insertion points for module pre- and post-ops.
        # Should roughly look so: first, determine subsets of nodes belonging to each
        # separate NNCF module (via scope analysis), then for each subset find input/output
        # edges using a corresponding NNCFGraph function; add a pre-op insertion point node as the
        # sink for input edges and connect it to input edge destinations, then add a post-op
        # insertion point as the source of output edges and connect it to output edge origins.

        node_keys_working_set = [deepcopy(node_key) for node_key in self.nodes.keys()]
        for operator_node_key in node_keys_working_set:
            original_node = self.nodes[operator_node_key][InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR]
            ia_op_exec_context = original_node[NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].input_agnostic

            # Pre-hook insertion point nodes
            pre_hook_insertion_point = InsertionPoint(ia_op_exec_context,
                                                      InsertionType.OPERATOR_PRE_HOOK)
            attrs = {
                InsertionPointGraph.NODE_TYPE_NODE_ATTR: InsertionPointGraphNodeType.INSERTION_POINT,
                InsertionPointGraph.INSERTION_POINT_DATA_NODE_ATTR: pre_hook_insertion_point,
            }
            ip_node_key = self.get_pre_hook_node_key(str(operator_node_key))
            self.add_node(ip_node_key, **attrs)
            in_edges = list(self.in_edges(operator_node_key))
            for from_node_key, to_node_key in in_edges:
                self.remove_edge(from_node_key, to_node_key)
                self.add_edge(from_node_key, ip_node_key)
            self.add_edge(ip_node_key, operator_node_key)
            operator_node = self.nodes[operator_node_key]
            operator_node[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR].add(ip_node_key)

            # Post-hook insertion point nodes
            post_hook_insertion_point = InsertionPoint(ia_op_exec_context,
                                                       InsertionType.OPERATOR_POST_HOOK)
            attrs = {
                InsertionPointGraph.NODE_TYPE_NODE_ATTR: InsertionPointGraphNodeType.INSERTION_POINT,
                InsertionPointGraph.INSERTION_POINT_DATA_NODE_ATTR: post_hook_insertion_point
            }
            ip_node_key = self.get_post_hook_node_key(str(operator_node_key))
            self.add_node(ip_node_key, **attrs)
            out_edges = list(self.out_edges(operator_node_key))
            for from_node_key, to_node_key in out_edges:
                self.remove_edge(from_node_key, to_node_key)
                self.add_edge(ip_node_key, to_node_key)
                # TODO: introduce separate insertion points for operator outputs if
                # the outputs are semantically different
            self.add_edge(operator_node_key, ip_node_key)
            operator_node = self.nodes[operator_node_key]
            operator_node[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR].add(ip_node_key)

    def _base_graph_match_has_breaking_output_edges(self, match):
        for node_key in match[:-1]:
            succs = list(self._base_nx_graph.succ[node_key].keys())
            for succ_key in succs:
                if succ_key not in match:
                    return True
        return False

    def get_ip_graph_with_merged_hw_optimized_operations(self,
                                                         hw_config: Optional[HWConfig] = None) -> 'InsertionPointGraph':
        #pylint:disable=too-many-branches
        merged_ip_graph = deepcopy(self)
        pattern = self._get_mergeable_operator_patterns(hw_config)
        from nncf.dynamic_graph.graph_matching import search_all
        matches = search_all(self._base_nx_graph, pattern)
        for match in matches:
            if len(match) == 1:
                continue

            input_node_key = match[0]
            output_node_key = match[-1]

            # If a subgraph has output edges in its middle, should skip merging it
            # Example:
            #       (conv2d)
            #          |------\
            #         (BN)    |
            #          |      |
            #        (RELU)   |
            #          |      |
            #        (cat)----/
            #          |
            #         ...

            has_breaking_output_edges = self._base_graph_match_has_breaking_output_edges(match)

            if has_breaking_output_edges:
                continue

            in_edges = list(self.in_edges(input_node_key))
            out_edges = list(self.out_edges(output_node_key))

            assert len(in_edges) <= 1  # TODO: change to == 1 when input nodes are handled correctly

            if in_edges:
                in_edge_key = in_edges[0]
                in_edge_copy = deepcopy(self.edges[in_edge_key])
            out_edge_copies_dict = {}
            for out_edge_key in out_edges:
                out_edge_copies_dict[out_edge_key] = deepcopy(self.edges[out_edge_key])

            conserved_edges_list = out_edges
            if in_edges:
                conserved_edges_list.append(in_edge_key)

            merged_node_attrs = deepcopy(self.nodes[input_node_key])
            merged_node_attrs[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR] = set()
            merged_node_key = ""
            for node_key in match:
                ip_node_keys = self.nodes[node_key][InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR]
                for ip_node_key in ip_node_keys:
                    should_keep_ip_node = False
                    for edge_key in conserved_edges_list:
                        if ip_node_key in edge_key:
                            should_keep_ip_node = True
                            break
                    if should_keep_ip_node:
                        merged_node_attrs[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR].add(ip_node_key)
                    else:
                        merged_ip_graph.remove_node(ip_node_key)
                merged_ip_graph.remove_node(node_key)
                merged_node_key += node_key + '\n'

            merged_ip_graph.add_node(merged_node_key, **merged_node_attrs)
            if in_edges:
                merged_ip_graph.add_edge(in_edge_key[0], merged_node_key, **in_edge_copy)
            for out_edge_key, out_edge_attrs in out_edge_copies_dict.items():
                merged_ip_graph.add_edge(merged_node_key, out_edge_key[1], **out_edge_attrs)

        return merged_ip_graph

    @staticmethod
    def get_pre_hook_node_key(node_key: str):
        return InsertionPointGraph.PRE_HOOK_ID_PREFIX + node_key

    @staticmethod
    def get_post_hook_node_key(node_key: str):
        return InsertionPointGraph.POST_HOOK_ID_PREFIX + node_key

    def _get_mergeable_operator_patterns(self, hw_config: Optional[HWConfig] = None) -> NodeExpression:
        """Resulting pattern should have single input; the operation with inputs to
        quantize should be the input operation; outputs should only be produced by one output node."""
        # TODO: Implement "repeating expressions" so that any number of "mergeable" operations
        # immediately following a linear/convolutional/matrix op are merged into one block
        import nncf.dynamic_graph.patterns as p
        pattern = p.LINEAR_OPS + p.ANY_BN_ACT_COMBO | p.LINEAR_OPS + p.ELTWISE_UNIFORM_OPS |\
                  p.ARITHMETIC + p.ANY_BN_ACT_COMBO | p.ANY_BN_ACT_COMBO
        return pattern

    def get_op_nodes_in_scope(self, scope: 'Scope') -> List:
        matching_ip_graph_op_nodes_list = []
        for node in self.nodes().values():
            if node[InsertionPointGraph.NODE_TYPE_NODE_ATTR] == InsertionPointGraphNodeType.OPERATOR:
                nncf_graph_node_ref = node[InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR]
                op_exec_context = nncf_graph_node_ref[NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR]
                op_scope = op_exec_context.input_agnostic.scope_in_model
                if op_scope in scope:
                    matching_ip_graph_op_nodes_list.append(node)
        return matching_ip_graph_op_nodes_list


# pylint: disable=too-many-public-methods


@ignore_scope
class NNCFNetwork(nn.Module, PostGraphBuildActing):
    def __init__(self, module, input_infos: List[ModelInputInfo] = None,
                 dummy_forward_fn=None, wrap_inputs_fn=None, scopes_without_shape_matching=None,
                 ignored_scopes=None, target_scopes=None):
        super().__init__()
        self._set_nncf_wrapped_model(module)
        self._forward_signature = inspect.signature(module.forward)
        self.input_infos = input_infos

        self.ignored_scopes = ignored_scopes
        self.target_scopes = target_scopes
        self._dummy_forward_fn = dummy_forward_fn

        device = next(module.parameters()).device

        if wrap_inputs_fn is not None:
            self._wrap_inputs_fn = wrap_inputs_fn
        else:
            self.__input_infos_based_input_wrapper = InputInfoWrapManager(self.input_infos,
                                                                          self._forward_signature,
                                                                          module_ref_for_device=self)
            self._wrap_inputs_fn = self.__input_infos_based_input_wrapper.wrap_inputs

        self._nncf_module_scopes = []  # type: List[Scope]
        self.scopes_without_shape_matching = scopes_without_shape_matching
        self.debug_interface = CombinedDebugInterface() if is_debug() else None
        self._extra_module_types = []  # type: List[CompressionModuleType]
        # pylint:disable=line-too-long
        self._insertions_into_original_graph = {}  # type: Dict[InsertionPoint, List[Tuple[Callable, OperationPriority]]]


        # all modules should be replaced prior to graph building
        self._replace_modules_by_nncf_modules(device)

        _orig_context = TracingContext()
        _orig_graph_build_forward_fn = self._get_dummy_forward_fn_for_graph_building(with_input_tracing=True)

        self._graph_builder = GraphBuilder(_orig_graph_build_forward_fn)

        _orig_context.add_node_comparators([MODEL_INPUT_OP_NAME], ShapeIgnoringTensorMetaComparator())
        if self.scopes_without_shape_matching:
            _orig_context.add_node_comparators(scopes_without_shape_matching,
                                               ShapeIgnoringTensorMetaComparator())

        self._original_graph = self._graph_builder.build_graph(self.get_nncf_wrapped_model(), _orig_context)

        self._compressed_context = TracingContext()

        self._dummy_forward_fn = self._get_dummy_forward_fn_for_graph_building(with_input_tracing=False)

        self._compressed_context.add_node_comparators([MODEL_INPUT_OP_NAME], ShapeIgnoringTensorMetaComparator())
        if self.scopes_without_shape_matching:
            self._compressed_context.add_node_comparators(scopes_without_shape_matching,
                                                          ShapeIgnoringTensorMetaComparator())
        self._load_listener = None

        self._builders = []  # type: List['CompressionAlgorithmBuilder']

    @debuggable_forward
    def forward(self, *args, **kwargs):
        with self._compressed_context as ctx:  # type: TracingContext
            ctx.base_module_thread_local_replica = self
            args, kwargs = self._wrap_inputs_fn(args, kwargs)
            retval = self.get_nncf_wrapped_model()(*args, **kwargs)
        return retval

    def register_algorithm(self, builder: 'CompressionAlgorithmBuilder'):
        """Should be called during *builder*'s *apply_to* method, otherwise there will be no corresponding
        controller returned by the network on the *commit_compression_changes* stage"""
        self._builders.append(builder)

    # Cannnot use property syntax here, otherwise the wrapped module will end up
    # being twice in the same checkpoint with different prefixes
    def get_nncf_wrapped_model(self):
        return getattr(self, MODEL_WRAPPED_BY_NNCF_ATTR_NAME)

    def _set_nncf_wrapped_model(self, value):
        setattr(self, MODEL_WRAPPED_BY_NNCF_ATTR_NAME, value)

    def get_modules_in_nncf_modules_by_type(self, types) -> Dict['Scope', nn.Module]:
        nncf_modules = self.get_nncf_modules()
        retval = {}
        for nncf_module_scope, nncf_module in nncf_modules.items():
            nncf_module_scope.pop()
            for relative_scope, target_module in get_all_modules_by_type(nncf_module, types).items():
                retval[nncf_module_scope + relative_scope] = target_module
        return retval

    def register_insertion_command(self, command: InsertionCommand):
        point = command.insertion_point
        if point not in self._insertions_into_original_graph:
            self._insertions_into_original_graph[point] = [(command.fn, command.priority)]
        else:
            self._insertions_into_original_graph[point].append((command.fn, command.priority))

    def commit_compression_changes(self) -> 'CompressionAlgorithmController':
        for insertion_point, fn_list_with_priority in self._insertions_into_original_graph.items():
            fn_list_with_priority = sorted(fn_list_with_priority, key=lambda x: x[1])
            self._insertions_into_original_graph[insertion_point] = fn_list_with_priority
            self._insert_at_point(insertion_point, [x[0] for x in fn_list_with_priority])

        if self.debug_interface is not None:
            self.debug_interface.init_actual(self)

        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        all_quantizations = get_state_dict_names_with_modules(self, quantization_types)
        self._load_listener = LoadStateListener(self, all_quantizations)

        if not self._builders:
            from nncf.algo_selector import NoCompressionAlgorithmController
            return NoCompressionAlgorithmController(self)

        if len(self._builders) == 1:
            return self._builders[0].build_controller(self)

        from nncf.composite_compression import CompositeCompressionAlgorithmController
        composite_controller = CompositeCompressionAlgorithmController(self)
        for algo_builder in self._builders:
            composite_controller.add(algo_builder.build_controller(self))
        return composite_controller

    def _insert_at_point(self, point: InsertionPoint, fn_list: List[Callable]):
        if point.insertion_type == InsertionType.OPERATOR_PRE_HOOK:
            self._compressed_context.register_pre_hooks(fn_list, point.ia_op_exec_context)
        elif point.insertion_type == InsertionType.OPERATOR_POST_HOOK:
            self._compressed_context.register_post_hooks(fn_list, point.ia_op_exec_context)
        else:
            norm_target_scope = self._normalize_variable_recurrent_scope(point.ia_op_exec_context.scope_in_model)
            norm_nncf_scopes = [self._normalize_variable_recurrent_scope(x) for x in self._nncf_module_scopes]
            assert norm_target_scope in norm_nncf_scopes  # Required for proper Recurrent/VariableRecurrent addressing
            nncf_module = self.get_module_by_scope(point.ia_op_exec_context.scope_in_model)
            if point.insertion_type == InsertionType.NNCF_MODULE_PRE_OP:
                for fn in fn_list:
                    nncf_module.register_pre_forward_operation(fn)
            elif point.insertion_type == InsertionType.NNCF_MODULE_POST_OP:
                for fn in fn_list:
                    nncf_module.register_post_forward_operation(fn)

    def __getattr__(self, name):
        wrapped_module = super().__getattr__(MODEL_WRAPPED_BY_NNCF_ATTR_NAME)
        if hasattr(wrapped_module, name):
            return getattr(wrapped_module, name)
        return super().__getattr__(name)

    def get_graph(self) -> NNCFGraph:
        return self._compressed_context.graph

    def get_original_graph(self) -> NNCFGraph:
        return self._original_graph

    def get_tracing_context(self) -> TracingContext:
        return self._compressed_context

    def _get_dummy_forward_fn_for_graph_building(self, with_input_tracing):
        if self._dummy_forward_fn is None:
            return create_dummy_forward_fn(self.input_infos,
                                           with_input_tracing=with_input_tracing,
                                           wrap_inputs_fn=self._wrap_inputs_fn)
        return self._dummy_forward_fn

    def _replace_modules_by_nncf_modules(self, device):
        module, self._nncf_module_scopes = replace_modules_by_nncf_modules(self.get_nncf_wrapped_model(),
                                                                           ignored_scopes=self.ignored_scopes,
                                                                           target_scopes=self.target_scopes)
        self._set_nncf_wrapped_model(module.to(device))

    def get_nncf_module_scopes(self) -> List['Scope']:
        return self._nncf_module_scopes

    def get_nncf_modules(self) -> Dict['Scope', torch.nn.Module]:
        nncf_module_names_list = NNCF_MODULES + [x.__name__ for x in NNCF_WRAPPED_USER_MODULES_DICT.values()]
        return get_all_modules_by_type(self.get_nncf_wrapped_model(), nncf_module_names_list)

    def rebuild_graph(self, *input_args):
        self._compressed_context.reset_graph()
        dummy_forward_fn = self._get_dummy_forward_fn_for_graph_building(with_input_tracing=False)
        builder = GraphBuilder(dummy_forward_fn)
        _ = builder.build_graph(self, self._compressed_context)

    def post_build_graph_actions(self):
        # Reset initialization flags (`initialized`) for all quantization modules
        # after dummy `load_state_dict` call.
        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        all_quantizations = get_state_dict_names_with_modules(self, quantization_types)
        for module in all_quantizations.values():
            module.initialized = False

    def get_post_pattern_insertion_points(self, pattern: 'NNCFNodeExpression',
                                          omit_nodes_in_nncf_modules=False) -> List[InsertionInfo]:
        io_infos = self._original_graph.get_matching_nncf_graph_pattern_io_list(pattern)

        insertion_infos = []
        for io_info in io_infos:
            # The input/output is given in terms of edges, but the post-hooks are currently applied to
            # nodes. Multiple output edges in a pattern I/O info may originate from one and the same
            # node, and we have to ensure that these resolve into just one insertion point - thus the usage of "set".
            pattern_insertion_info_set = set()
            if len(io_info.output_edges) > 1:
                nncf_logger.debug("WARNING: pattern has more than one activation output")

            for nncf_node in io_info.output_nodes:
                pattern_insertion_info_set.add(InsertionInfo(nncf_node.op_exec_context,
                                                             is_output=True,
                                                             shape_to_operate_on=None))
                # TODO: determine output shapes for output nodes to enable per-channel quantization

            # Ignore input nodes in the pattern for now, rely on the _quantize_inputs functions.
            # TODO: handle input quantization here as well

            # Since this function is currently only used for activation quantization purposes via operator
            # post-hook mechanism, we may take any edge and it will point from the same node where we will have to
            # insert a quantizer later. However, in the future the output edges may refer to activation tensors
            # with different sizes, in which case we have to insert different per-channel quantizers to
            # accomodate different trainable params if there is a difference in the channel dimension.
            # Furthermore, currently there is no distinction for single tensor output to multiple nodes and
            # multiple tensor output to multiple nodes ("chunk" operation is an example of the latter).
            # The pattern may also have unexpected outputs from a node in the middle of the pattern (see
            # "densenet121.dot" for an example of this) - need to decide what to do with that in terms
            # of quantization.
            # TODO: address the issues above.

            for nncf_edge in io_info.output_edges:
                pattern_insertion_info_set.add(InsertionInfo(nncf_edge.from_node.op_exec_context,
                                                             is_output=False,
                                                             shape_to_operate_on=nncf_edge.tensor_shape))
            insertion_infos += list(pattern_insertion_info_set)

        insertion_infos = list(
            set(insertion_infos))  # Filter the overlapping insertion points from different matches (happens for GNMT)
        insertion_infos_filtered = []

        for info in insertion_infos:
            if omit_nodes_in_nncf_modules and self.is_scope_in_nncf_module_scope(info.op_exec_context.scope_in_model):
                continue
            insertion_infos_filtered.append(info)

        return insertion_infos_filtered

    def is_scope_in_nncf_module_scope(self, scope: 'Scope'):
        # TODO: optimize
        norm_nncf_scopes = [self._normalize_variable_recurrent_scope(x) for x in self._nncf_module_scopes]
        norm_op_scope = self._normalize_variable_recurrent_scope(scope)
        for nncf_scope in norm_nncf_scopes:
            if norm_op_scope in nncf_scope:
                return True
        return False

    def register_compression_module_type(self, compression_module_type: CompressionModuleType):
        attr_name = self._compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type in self._extra_module_types:
            raise RuntimeError("Module type {} is already registered".format(compression_module_type))
        self.__setattr__(attr_name, nn.ModuleDict())
        self._extra_module_types.append(compression_module_type)

    def add_compression_module(self, module_key: str, module: nn.Module,
                               compression_module_type: CompressionModuleType):
        attr_name = self._compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type not in self._extra_module_types:
            raise RuntimeError("Module type {} was not registered".format(compression_module_type))
        self.__getattr__(attr_name)[module_key] = module

    def get_compression_modules_by_type(self, compression_module_type: CompressionModuleType) -> nn.ModuleDict:
        attr_name = self._compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type not in self._extra_module_types:
            raise RuntimeError("Module type {} was not registered".format(compression_module_type))
        return self.__getattr__(attr_name)

    @staticmethod
    def _compression_module_type_to_attr_name(compression_module_type: CompressionModuleType):
        """Required for backward compatibility with checkpoints that store function and activation
        quantizers directly under corresponding attributes of NNCFNetwork."""
        if compression_module_type == CompressionModuleType.FUNCTION_QUANTIZER:
            return "function_quantizers"
        if compression_module_type == CompressionModuleType.ACTIVATION_QUANTIZER:
            return "activation_quantizers"
        raise RuntimeError("Unknown extra module type")

    def sort_compression_modules(self, compression_module_type: CompressionModuleType):
        attr_name = self._compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type not in self._extra_module_types:
            raise RuntimeError("Module type {} was not registered".format(compression_module_type))
        module_dict = self.__getattr__(attr_name)
        # pylint: disable=protected-access
        module_dict._modules = OrderedDict(sorted(module_dict._modules.items()))
        self.__setattr__(attr_name, module_dict)

    @staticmethod
    def _normalize_variable_recurrent_scope(scope: 'Scope'):
        """
        Two scopes pointing to an NNCF module that only differ in a Recurrent/VariableRecurrent/VariableRecurrentReverse
        scope element actually point to one and the same module.
        """
        ret_scope = scope.copy()
        for scope_element in ret_scope:
            if scope_element.calling_module_class_name in ["Recurrent", "VariableRecurrent",
                                                           "VariableRecurrentReverse"]:
                scope_element.calling_module_class_name = "NormalizedName_Recurrent"
        return ret_scope

    def do_dummy_forward(self, force_eval=False):
        """Attention: If run with force_eval=False, this may spoil the batchnorm statistics,
        and an eval run of the model will perform much worse than the train run. """
        if force_eval:
            train_mode = self.training
            self.eval()
        with torch.no_grad():
            self._dummy_forward_fn(self)
        if force_eval:
            if train_mode:
                self.train()

    def get_insertion_point_graph(self) -> InsertionPointGraph:
        ip_graph = InsertionPointGraph(self._original_graph.get_nx_graph_copy())

        # Mark IP graph operator nodes with associated op metatypes
        # Determining operator metatypes is more suited to occur at wrap_operator
        # stage, because it might be influenced by specific non-tensor function paramters,
        # but we have to inspect the containing module parameters as well, so the
        # TracingContext in wrap_operator would have to retain a reference to
        # the model that uses it. Since currently we do not need to inspect the
        # function arguments to determine the metatype, we can do this here, but
        # once we need to inspect the arguments, the code will have to be moved to
        # wrap_operator.

        for node_key in ip_graph.nodes:
            ip_graph_node = ip_graph.nodes[node_key]
            ip_graph_node_type = ip_graph_node[InsertionPointGraph.NODE_TYPE_NODE_ATTR]
            if ip_graph_node_type == InsertionPointGraphNodeType.OPERATOR:
                nncf_graph_node_ref = ip_graph_node[InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR]
                op_exec_context = nncf_graph_node_ref[NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR]
                op_name = op_exec_context.operator_name
                scope = op_exec_context.scope_in_model
                op_arch = OPERATOR_METATYPES.get_operator_metatype_by_op_name(op_name)
                module = self.get_module_by_scope(scope)
                if module is not None:
                    subtype = op_arch.determine_subtype(containing_module=module)
                    if subtype is not None:
                        op_arch = subtype
                ip_graph_node[InsertionPointGraph.OPERATOR_METATYPE_NODE_ATTR] = op_arch
        return ip_graph

    def get_module_by_scope(self, scope: 'Scope') -> torch.nn.Module:
        curr_module = self.get_nncf_wrapped_model()
        for scope_element in scope[1:]:  # omit first scope element which corresponds to base module
            if scope_element.calling_field_name is None:
                # The module used is being created in-place every time and never stored in the model,
                # happens for nn.Softmax in BERT implementations.
                return None
            # pylint: disable=protected-access
            next_module = curr_module._modules.get(scope_element.calling_field_name)
            if next_module is None:
                raise RuntimeError("Could not find a {} module member in {} module of scope {} during node search"
                                   .format(scope_element.calling_field_name,
                                           scope_element.calling_module_class_name,
                                           str(scope)))
            curr_module = next_module
        return curr_module

    def get_parameters_count_in_model(self):
        """
        Return total amount of model parameters.
        """
        count = 0
        for param in self.parameters():
            count = count + param.numel()
        return count

    def get_flops_per_module(self):
        """
        Calculates FLOPS count for modules.
        """
        model = self
        flops_count_dict = {}

        def get_hook(name):
            def compute_MACs_hook(module, input_, output):
                if isinstance(module, (nn.Conv2d, nn.ConvTranspose2d)):
                    ks = module.weight.data.shape
                    mac_count = ks[0] * ks[1] * ks[2] * ks[3] * output.shape[3] * output.shape[2]
                elif isinstance(module, nn.Linear):
                    mac_count = input_[0].shape[1] * output.shape[-1]
                elif isinstance(module, nn.BatchNorm2d):
                    mac_count = np.prod(list(input_[0].shape))
                else:
                    return
                flops_count_dict[name] = 2 * mac_count

            return compute_MACs_hook

        hook_list = [m.register_forward_hook(get_hook(n)) for n, m in model.named_modules()]

        model.do_dummy_forward(force_eval=True)

        for h in hook_list:
            h.remove()
        return flops_count_dict

    def get_MACs_in_model(self):
        """
            Calculates MAC units count for model.
        """
        flops_count_dict = self.get_flops_per_module()
        total_MACs_count = sum(v // 2 for v in flops_count_dict.values())
        return total_MACs_count

    def get_input_infos(self) -> List[ModelInputInfo]:
        return deepcopy(self.input_infos)



"""
 Copyright (c) 2019-2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from collections import namedtuple
from typing import List

from texttable import Texttable

from nncf.compression_method_api import CompressionAlgorithmBuilder, CompressionAlgorithmController, CompressionLevel
from nncf.dynamic_graph.graph import InputAgnosticOperationExecutionContext
from nncf.layer_utils import COMPRESSION_MODULES
from nncf.module_operations import UpdateWeight
from nncf.nncf_logger import logger as nncf_logger
from nncf.nncf_network import InsertionCommand, InsertionPoint, InsertionType, OperationPriority
from nncf.nncf_network import NNCFNetwork

SparseModuleInfo = namedtuple('SparseModuleInfo', ['module_name', 'module', 'operand'])


class BaseSparsityAlgoBuilder(CompressionAlgorithmBuilder):
    def __init__(self, config, should_init: bool = True):
        super().__init__(config, should_init)
        self._sparsified_module_info = []

    def apply_to(self, target_model: NNCFNetwork) -> NNCFNetwork:
        insertion_commands = self._sparsify_weights(target_model)
        for command in insertion_commands:
            target_model.register_insertion_command(command)
        target_model.register_algorithm(self)
        return target_model

    def _sparsify_weights(self, target_model: NNCFNetwork) -> List[InsertionCommand]:
        device = next(target_model.parameters()).device
        sparsified_modules = target_model.get_nncf_modules()
        insertion_commands = []
        for module_scope, module in sparsified_modules.items():
            scope_str = str(module_scope)

            if not self._should_consider_scope(scope_str):
                nncf_logger.info("Ignored adding Weight Sparsifier in scope: {}".format(scope_str))
                continue

            nncf_logger.info("Adding Weight Sparsifier in scope: {}".format(scope_str))
            operation = self.create_weight_sparsifying_operation(module)
            hook = UpdateWeight(operation).to(device)
            insertion_commands.append(InsertionCommand(InsertionPoint(
                InputAgnosticOperationExecutionContext("", module_scope, 0),
                InsertionType.NNCF_MODULE_PRE_OP), hook, OperationPriority.SPARSIFICATION_PRIORITY))
            self._sparsified_module_info.append(
                SparseModuleInfo(scope_str, module, hook.operand))

        return insertion_commands

    def build_controller(self, target_model: NNCFNetwork) -> CompressionAlgorithmController:
        return BaseSparsityAlgoController(target_model, self._sparsified_module_info)

    def create_weight_sparsifying_operation(self, target_module):
        raise NotImplementedError


class BaseSparsityAlgoController(CompressionAlgorithmController):
    def __init__(self, target_model: NNCFNetwork,
                 sparsified_module_info: List[SparseModuleInfo]):
        super().__init__(target_model)
        self.sparsified_module_info = sparsified_module_info

    def freeze(self):
        raise NotImplementedError

    def set_sparsity_level(self, sparsity_level: float):
        raise NotImplementedError

    @property
    def sparsified_weights_count(self):
        count = 0
        for minfo in self.sparsified_module_info:
            count = count + minfo.module.weight.view(-1).size(0)
        return max(count, 1)

    @property
    def sparsity_rate_for_sparsified_modules(self):
        nonzero = 0
        count = 0

        for minfo in self.sparsified_module_info:
            mask = minfo.operand.apply_binary_mask(minfo.module.weight)
            nonzero = nonzero + mask.nonzero().size(0)
            count = count + mask.view(-1).size(0)

        return 1 - nonzero / max(count, 1)

    @property
    def sparsity_rate_for_model(self):
        nonzero = 0
        count = 0

        for m in self._model.modules():
            if isinstance(m, tuple(COMPRESSION_MODULES.registry_dict.values())):
                continue

            sparsified_module = False
            for minfo in self.sparsified_module_info:
                if minfo.module == m:
                    mask = minfo.operand.apply_binary_mask(m.weight)
                    nonzero = nonzero + mask.nonzero().size(0)
                    count = count + mask.numel()

                    if not m.bias is None:
                        nonzero = nonzero + m.bias.nonzero().size(0)
                        count = count + m.bias.numel()

                    sparsified_module = True

            if not sparsified_module:
                for param in m.parameters(recurse=False):
                    nonzero = nonzero + param.nonzero().size(0)
                    count = count + param.numel()

        return 1 - nonzero / max(count, 1)

    def statistics(self):
        stats = super().statistics()
        table = Texttable()
        header = ["Name", "Weight's Shape", "SR", "% weights"]
        data = [header]

        sparsified_weights_count = self.sparsified_weights_count

        for minfo in self.sparsified_module_info:
            drow = {h: 0 for h in header}
            drow["Name"] = minfo.module_name
            drow["Weight's Shape"] = list(minfo.module.weight.size())
            mask = minfo.operand.apply_binary_mask(minfo.module.weight)
            nonzero = mask.nonzero().size(0)
            drow["SR"] = 1.0 - nonzero / max(mask.view(-1).size(0), 1)
            drow["% weights"] = (mask.view(-1).size(0) / sparsified_weights_count) * 100
            row = [drow[h] for h in header]
            data.append(row)
        table.add_rows(data)

        stats["sparsity_statistic_by_module"] = table
        stats["sparsity_rate_for_sparsified_modules"] = self.sparsity_rate_for_sparsified_modules
        stats["sparsity_rate_for_model"] = self.sparsity_rate_for_model

        return self.add_algo_specific_stats(stats)

    def add_algo_specific_stats(self, stats):
        return stats

    def compression_level(self) -> CompressionLevel:
        return CompressionLevel.FULL



from .precision_init.hawq_init import HAWQPrecisionInitializer
from .precision_init.manual_init import ManualPrecisionInitializer


class PrecisionInitializerFactory:
    @staticmethod
    def create(init_type: str):
        if init_type == "manual":
            return ManualPrecisionInitializer
        if init_type == "hawq":
            return HAWQPrecisionInitializer
        raise NotImplementedError



"""
 Copyright (c) 2019-2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from collections import OrderedDict
from typing import Callable, Any, List, Optional
from copy import deepcopy

import torch



class ModelInputInfo:
    FILLER_TYPE_ONES = "ones"
    FILLER_TYPE_ZEROS = "zeros"
    FILLER_TYPE_RANDOM = "random"
    FILLER_TYPES = [FILLER_TYPE_ONES, FILLER_TYPE_ZEROS, FILLER_TYPE_RANDOM]

    def __init__(self, shape: List[int], type_str: str = "float", keyword=None, filler=None):
        self.shape = shape
        self.type = self._string_to_torch_type(type_str)
        self.keyword = keyword
        if filler is None:
            self.filler = self.FILLER_TYPE_ONES
        else:
            self.filler = filler
            if self.filler not in self.FILLER_TYPES:
                raise RuntimeError("Unknown input filler type: {}".format(filler))

    @staticmethod
    def _string_to_torch_type(string):
        if string == "long":
            return torch.long
        return torch.float32

    @staticmethod
    def torch_type_to_string(dtype: torch.dtype):
        if dtype is torch.long:
            return "long"
        return "float"

    def is_integer_input(self):
        return self.type != torch.float32

    def __eq__(self, other):
        return self.type == other.type and self.keyword == other.keyword


def create_input_infos(config) -> List[ModelInputInfo]:
    input_infos = config.get("input_info", [])
    if isinstance(input_infos, dict):
        return [ModelInputInfo(input_infos.get("sample_size"),
                               input_infos.get("type"),
                               input_infos.get("keyword"),
                               input_infos.get("filler")), ]
    if isinstance(input_infos, list):
        if not input_infos:
            return [ModelInputInfo([1, 3, 224, 224])]
        return [ModelInputInfo(info_dict.get("sample_size"),
                               info_dict.get("type"),
                               info_dict.get("keyword"),
                               info_dict.get("filler")) for info_dict in input_infos]
    raise RuntimeError("Invalid input_infos specified in config - should be either dict or list of dicts")


def create_mock_tensor(input_info: ModelInputInfo, device: str):
    args = {"size": input_info.shape, "dtype": input_info.type, "device": device}
    if input_info.filler == ModelInputInfo.FILLER_TYPE_ZEROS:
        return torch.zeros(**args)
    if input_info.filler == ModelInputInfo.FILLER_TYPE_ONES:
        return torch.ones(**args)
    if input_info.filler == ModelInputInfo.FILLER_TYPE_RANDOM:
        return torch.rand(**args)
    raise RuntimeError


class GraphBuilder:
    def __init__(self, custom_forward_fn: Callable[[torch.nn.Module], Any]):
        self.custom_forward_fn = custom_forward_fn

    def build_graph(self, model: torch.nn.Module, context_to_use: Optional['TracingContext'] = None) -> 'NNCFGraph':
        sd = deepcopy(model.state_dict())

        from nncf.dynamic_graph.context import TracingContext
        if context_to_use is None:
            context_to_use = TracingContext()

        context_to_use.base_module_thread_local_replica = model
        with context_to_use as _ctx:
            with torch.no_grad():
                self.custom_forward_fn(model)
        model.load_state_dict(sd)

        if isinstance(model, PostGraphBuildActing):
            model.post_build_graph_actions()
        return context_to_use.graph


class PostGraphBuildActing:
    def post_build_graph_actions(self):
        pass


def create_dummy_forward_fn(input_infos: List[ModelInputInfo], with_input_tracing=False,
                            wrap_inputs_fn=None):
    from nncf.dynamic_graph.input_wrapping import wrap_nncf_model_inputs_with_objwalk

    def default_dummy_forward_fn(model):
        device = next(model.parameters()).device
        args_list = [create_mock_tensor(info, device) for info in input_infos if info.keyword is None]
        kwargs = OrderedDict()
        for info in input_infos:
            if info.keyword is not None:
                kwargs[info.keyword] = create_mock_tensor(info, device)
        args = tuple(args_list)

        if with_input_tracing:
            if wrap_inputs_fn is None:
                # We control the input argument structure w.r.t. tensors
                # - a simple objwalk application should be sufficient in this simple case.
                # For more control, wrap_inputs_fn is used when this is used in NNCFNetwork
                # which is guaranteed to be the same as during the actual NNCFNetwork.forward
                args, kwargs = wrap_nncf_model_inputs_with_objwalk(args, kwargs)
            else:
                args, kwargs = wrap_inputs_fn(args, kwargs)

        return model(*args, **kwargs)

    return default_dummy_forward_fn



"""
 Copyright (c) 2019-2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint:disable=too-many-lines
from collections import deque
from enum import Enum
from typing import Dict, Tuple, Set, Any, Callable

import networkx as nx
import warnings
from copy import deepcopy

from nncf.dynamic_graph.graph import OperationExecutionContext, NNCFGraph, InputAgnosticOperationExecutionContext
# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
from nncf.dynamic_graph.graph_builder import ModelInputInfo
from nncf.dynamic_graph.operator_metatypes import *
from nncf.dynamic_graph.operator_metatypes import OPERATOR_METATYPES
from nncf.hw_config import HWConfig
from nncf.dynamic_graph.input_wrapping import MODEL_INPUT_OP_NAME
from nncf.nncf_network import InsertionInfo, InsertionType, InsertionPointGraph, InsertionPointGraphNodeType, \
    InsertionPoint
from nncf.quantization.layers import QuantizerConfig, QuantizationMode
from nncf.utils import in_scope_list
from nncf.nncf_logger import logger as nncf_logger


class QuantizationTrait(Enum):
    """General, hardware-agnostic specifications for the relation of operators to quantization.
    Hardware-specific quantization configuration is handled elsewhere."""
    NON_QUANTIZABLE = -1
    QUANTIZATION_AGNOSTIC = 0
    INPUTS_QUANTIZABLE = 1


DEFAULT_QUANT_TRAIT_TO_OP_DICT = {
    QuantizationTrait.INPUTS_QUANTIZABLE: [
        Conv2dMetatype,
        Conv3dMetatype,
        ConvTranspose2dMetatype,
        ConvTranspose3dMetatype,
        DepthwiseConv2dSubtype,
        DepthwiseConv3dSubtype,
        LinearMetatype,
        HardTanhMetatype,
        TanhMetatype,
        ELUMetatype,
        PRELUMetatype,
        LayerNormMetatype,
        GELUMetatype,
        SigmoidMetatype,
        AddMetatype,
        MulMetatype,
        DivMetatype,
        ExpMetatype,
        ErfMetatype,
        MatMulMetatype,
        MeanMetatype,
        RoundMetatype,
        PixelShuffleMetatype,
        BatchNormMetatype
    ],
    QuantizationTrait.NON_QUANTIZABLE: [
        EmbeddingMetatype,
        SoftmaxMetatype
    ],
}  # type: Dict[QuantizationTrait, List[OperatorMetatype]]


class QuantizerInsertionInfo(InsertionInfo):
    def __init__(self, op_exec_context: OperationExecutionContext, is_input=False, is_output=False,
                 shape_to_operate_on=None,
                 quantizers_between_quantizable_layers: 'QuantizersBetweenQuantizableLayers' = None):
        super().__init__(op_exec_context, is_input, is_output, shape_to_operate_on)
        self.quantizers_between_quantizable_layers = quantizers_between_quantizable_layers

    @staticmethod
    def from_insertion_info(insertion_info: InsertionInfo) -> 'QuantizerInsertionInfo':
        result = QuantizerInsertionInfo(op_exec_context=insertion_info.op_exec_context,
                                        is_input=insertion_info.is_input,
                                        is_output=insertion_info.is_output,
                                        shape_to_operate_on=insertion_info.shape_to_operate_on)
        result.linked_op_exec_contexts = insertion_info.linked_op_exec_contexts
        return result

class PropagatingQuantizer:
    """Used in conjunction with QuantizerPropagationStateGraph to keep track of
       the allowed quantization configs corresponding to the model operation node
       whose inputs it quantizes, and also of the nodes/edges in the model control
       graph that this quantizer affects. It should be moved against the data flow of
       the model, tracking the affected nodes and edges of
       QuantizerPropagationStateGraph. No actual quantization modules are used here,
       only the associated configs (such as bitwidths, modes, signed/unsigned
       attributes etc.)"""

    def __init__(self, id_: int, quant_configs: List[QuantizerConfig], init_location_node_key: str,
                 unified_scale: bool = False):
        self._potential_quant_configs = quant_configs  # type: List[QuantizerConfig]
        self.affected_edges = set()
        self.affected_ip_nodes = set()
        self.propagation_path = []
        self.current_location_node_key = init_location_node_key
        self.last_accepting_location_node_key = None
        self.id = id_
        self.unified_scale = unified_scale

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def potential_quant_configs(self) -> List[QuantizerConfig]:
        return self._potential_quant_configs


class TransitionStatus(Enum):
    SHOULD_TRANSITION = 0
    SHOULD_MERGE = 1
    SHOULD_NOT_TRANSITION = 2


class PropagationStrategy(Enum):
    CONSERVATIVE = 0  # While propagating up through a downward-branching node,
                      # do not propagate if the propagation results in narrowing the list of
                      # quantization variants available to quantizers on neighbouring branches
    AGGRESSIVE = 1


class QuantizerPropagationStateGraphNodeType(Enum):
    INSERTION_POINT = 0
    OPERATOR = 1
    AUXILIARY_BARRIER = 2


class UnifiedScalePropagatingQuantizerGroupManager:
    def __init__(self):
        self._next_gid = 0
        self._group_vs_prop_quants_dict = {}  # type: Dict[int, Set[PropagatingQuantizer]]

    def _get_next_gid(self) -> int:
        retval = self._next_gid
        self._next_gid += 1
        return retval

    def register_group(self, prop_quants: Set[PropagatingQuantizer]) -> int:
        for pq in prop_quants:
            for gid, group in self._group_vs_prop_quants_dict.items():
                assert pq not in group, "Propagating quantizer #{} is already registered in a group {}!".format(pq.id,
                                                                                                                gid)
        gid = self._get_next_gid()
        self._group_vs_prop_quants_dict[gid] = prop_quants
        return gid

    def add_to_group(self, target_gid: int, prop_quant: PropagatingQuantizer):
        for gid, group in self._group_vs_prop_quants_dict.items():
            if target_gid != gid:
                assert prop_quant not in group, "Tried to add propagating quantizer #{} to group #{}, " \
                                                "but it is already registered in a group {}!".format(prop_quant.id,
                                                                                                     target_gid,
                                                                                                     gid)
        self._group_vs_prop_quants_dict[target_gid].add(prop_quant)

    def remove_from_group(self, group: int, prop_quant: PropagatingQuantizer):
        self._group_vs_prop_quants_dict[group].remove(prop_quant)

    def get_group_vs_prop_quants_dict(self) -> Dict[int, Set[PropagatingQuantizer]]:
        return copy(self._group_vs_prop_quants_dict)

    def get_group_id_by_propagating_quantizer_id(self, requested_pqid: int) -> Optional[int]:
        for gid, group in self._group_vs_prop_quants_dict.items():
            for pq in group:
                if pq.id == requested_pqid:
                    return gid
        return None

    def merge_groups(self, merge_to_gid: int, merge_from_gid: int):
        if merge_to_gid == merge_from_gid:
            return
        self._group_vs_prop_quants_dict[merge_to_gid].update(self._group_vs_prop_quants_dict[merge_from_gid])
        self._group_vs_prop_quants_dict.pop(merge_from_gid)


class QuantizerPropagationStateGraph(nx.DiGraph):
    """This class is based upon InsertionPointGraph and represents
       a"chessboard" for PropagatingQuantizer items.  It tracks the current state of
       quantizer propagation by associating the operator and insertion point nodes and
       edges to propagating quantizers, if any. It can move a propagating quantizer
       via own edges and mark its progress through the graph, which is required for
       resolving situations when multiple quantizers attempt to proceed via one and
       the same graph node/edge. This class is mainly operated upon by the
       QuantizerPropagationSolver objects."""
    PROPAGATING_QUANTIZER_NODE_ATTR = "propagating_quantizer"
    AFFECTING_PROPAGATING_QUANTIZERS_ATTR = "affecting_propagating_quantizers"
    QUANTIZATION_TRAIT_NODE_ATTR = "quantization_trait"
    ALLOWED_INPUT_QUANTIZATION_TYPES_NODE_ATTR = "allowed_input_quantization_types"
    OPERATOR_METATYPE_NODE_ATTR = "op_meta"
    OPERATOR_SCOPE = "op_scope"
    INSERTION_POINT_DATA_NODE_ATTR = "insertion_point"
    NODE_TYPE_NODE_ATTR = "node_type"
    BARRIER_NODE_KEY_POSTFIX = "BARRIER"

    def __init__(self, ip_graph: InsertionPointGraph, ignored_scopes=None):
        super().__init__()
        ip_graph = deepcopy(ip_graph)
        self._created_prop_quantizer_counter = 0

        self._ignored_scopes = deepcopy(ignored_scopes)
        self.ignored_node_keys = []

        self._unified_scale_group_manager = UnifiedScalePropagatingQuantizerGroupManager()
        self._input_node_keys_vs_contexts = {}  # type: Dict[str, InputAgnosticOperationExecutionContext]

        barrier_node_extra_edges = []
        for node_key, node in ip_graph.nodes.items():
            qpg_node = {
                self.NODE_TYPE_NODE_ATTR: \
                    self.ipg_node_type_to_qpsg_node_type(node[InsertionPointGraph.NODE_TYPE_NODE_ATTR])}
            if node[InsertionPointGraph.NODE_TYPE_NODE_ATTR] == InsertionPointGraphNodeType.INSERTION_POINT:
                qpg_node[self.PROPAGATING_QUANTIZER_NODE_ATTR] = None
                qpg_node[self.AFFECTING_PROPAGATING_QUANTIZERS_ATTR] = []
                qpg_node[self.INSERTION_POINT_DATA_NODE_ATTR] = node[
                    InsertionPointGraph.INSERTION_POINT_DATA_NODE_ATTR]
            elif node[InsertionPointGraph.NODE_TYPE_NODE_ATTR] == InsertionPointGraphNodeType.OPERATOR:
                qpg_node[self.ALLOWED_INPUT_QUANTIZATION_TYPES_NODE_ATTR] = set()
                qpg_node[
                    self.QUANTIZATION_TRAIT_NODE_ATTR] = QuantizationTrait.NON_QUANTIZABLE
                qpg_node[self.AFFECTING_PROPAGATING_QUANTIZERS_ATTR] = []
                qpg_node[self.OPERATOR_METATYPE_NODE_ATTR] = node[InsertionPointGraph.OPERATOR_METATYPE_NODE_ATTR]
                node_ia_op_exec_context = node[InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR][
                    NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].input_agnostic  # type: InputAgnosticOperationExecutionContext
                qpg_node[self.OPERATOR_SCOPE] = node_ia_op_exec_context.scope_in_model
                node_scope = str(node_ia_op_exec_context)

                op_name = node_ia_op_exec_context.operator_name
                if op_name == MODEL_INPUT_OP_NAME:
                    self._input_node_keys_vs_contexts[node_key] = node_ia_op_exec_context

                if in_scope_list(node_scope, self._ignored_scopes):
                    self.ignored_node_keys.append(node_key)
                    qpg_node_barrier = {
                        self.NODE_TYPE_NODE_ATTR: QuantizerPropagationStateGraphNodeType.AUXILIARY_BARRIER,
                        'label': QuantizerPropagationStateGraph.BARRIER_NODE_KEY_POSTFIX}
                    barrier_node_key = self.get_barrier_node_key(node_key)
                    self.add_node(barrier_node_key, **qpg_node_barrier)
                    barrier_node_extra_edges.append((barrier_node_key, node_key))

            self.add_node(node_key, **qpg_node)

        for from_node, to_node, edge_data in ip_graph.edges(data=True):
            edge_data[self.AFFECTING_PROPAGATING_QUANTIZERS_ATTR] = []
            self.add_edge(from_node, to_node, **edge_data)

        for u_node_key, v_node_key in barrier_node_extra_edges:
            edge_attr = {QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR: []}
            next_v_node_key = list(self.succ[v_node_key].keys())[0]  # POST HOOK v
            self.add_edge(v_node_key, u_node_key, **edge_attr)
            self.add_edge(u_node_key, next_v_node_key, **edge_attr)
            self.remove_edge(v_node_key, next_v_node_key)

    @staticmethod
    def ipg_node_type_to_qpsg_node_type(ipg_node_type: InsertionPointGraphNodeType) \
        -> QuantizerPropagationStateGraphNodeType:
        if ipg_node_type == InsertionPointGraphNodeType.INSERTION_POINT:
            return QuantizerPropagationStateGraphNodeType.INSERTION_POINT
        if ipg_node_type == InsertionPointGraphNodeType.OPERATOR:
            return QuantizerPropagationStateGraphNodeType.OPERATOR
        raise RuntimeError("Invalid insertion point graph node type.")

    @staticmethod
    def get_barrier_node_key(node_key: str):
        return QuantizerPropagationStateGraph.BARRIER_NODE_KEY_POSTFIX + node_key

    # pylint:disable=too-many-branches
    def merge_quantizer_into_path(self, prop_quantizer: PropagatingQuantizer, path: List):
        curr_node = self.nodes[prop_quantizer.current_location_node_key]
        curr_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = None
        surviving_quantizers = []  # type: List[PropagatingQuantizer]
        for from_node_key, to_node_key in path:
            edge = self.edges[from_node_key, to_node_key]
            potential_quantizers = edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
            if potential_quantizers:
                surviving_quantizers = potential_quantizers
                break
            from_node = self.nodes[from_node_key]
            potential_quantizer = from_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR]
            if potential_quantizer is None:
                if from_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]:
                    potential_quantizer = \
                        from_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR][0]

            if potential_quantizer is not None:
                prop_quantizer.affected_edges.add((from_node_key, to_node_key))
                edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(prop_quantizer)
                surviving_quantizers.append(potential_quantizer)
                break

        if surviving_quantizers:
            for pq in surviving_quantizers:
                pq.affected_edges.update(prop_quantizer.affected_edges)
                for from_node_key, to_node_key in prop_quantizer.affected_edges:
                    from_node = self.nodes[from_node_key]
                    from_node_type = from_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
                    if from_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                        # pylint:disable=line-too-long
                        self.nodes[from_node_key][QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(pq)
            if prop_quantizer.unified_scale:
                gid = self._unified_scale_group_manager.get_group_id_by_propagating_quantizer_id(prop_quantizer.id)
                for other_pq in surviving_quantizers:
                    if other_pq.unified_scale:
                        other_gid = self._unified_scale_group_manager.get_group_id_by_propagating_quantizer_id(
                            other_pq.id)
                        self._unified_scale_group_manager.merge_groups(gid, other_gid)
                    else:
                        self._unified_scale_group_manager.add_to_group(gid, other_pq)

            for affected_edge_tuple in prop_quantizer.affected_edges:
                edge = self.edges[affected_edge_tuple]
                affecting_quantizers = edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
                for pq in surviving_quantizers:
                    affecting_quantizers.append(pq)
            self.remove_propagating_quantizer(prop_quantizer)
        else:
            raise RuntimeError("Surviving_quantizers not found !"
                               " Nodes quantized with quantizer #{} will be lost".format(prop_quantizer.id))

    def backtrack_propagation_until_accepting_location(self, prop_quantizer: PropagatingQuantizer) -> Optional[
            PropagatingQuantizer]:
        if prop_quantizer.last_accepting_location_node_key is None:
            # The quantizer was stillborn.
            # If there are quantizer-affected inbound edges, should transfer this quantizer's
            # affected edges and nodes to the inbound edge quantizers
            curr_node_key = prop_quantizer.current_location_node_key
            inbound_affecting_quantizers = set()
            for in_edge_key in self.in_edges(curr_node_key):
                in_edge = self.edges[in_edge_key]
                inbound_affecting_quantizers.update(
                    in_edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR])

            for inbound_pq in inbound_affecting_quantizers:
                inbound_pq.affected_edges.update(prop_quantizer.affected_edges)
                inbound_pq.affected_ip_nodes.update(prop_quantizer.affected_ip_nodes)
            for edge in prop_quantizer.affected_edges:
                self.edges[edge][QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR] += list(
                    inbound_affecting_quantizers)
            for ip_node_key in prop_quantizer.affected_ip_nodes:
                self.nodes[ip_node_key][QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR] += list(
                    inbound_affecting_quantizers)

            self.remove_propagating_quantizer(prop_quantizer)
            return None

        curr_node_key = prop_quantizer.current_location_node_key
        curr_node = self.nodes[curr_node_key]
        curr_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = None
        while prop_quantizer.current_location_node_key != prop_quantizer.last_accepting_location_node_key:
            from_node_key, to_node_key = prop_quantizer.propagation_path.pop()

            edge = self.edges[from_node_key, to_node_key]
            edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].remove(prop_quantizer)
            prop_quantizer.affected_edges.remove((from_node_key, to_node_key))
            from_node = self.nodes[from_node_key]
            from_node_type = from_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if from_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                from_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].remove(prop_quantizer)
                prop_quantizer.affected_ip_nodes.remove(from_node_key)

            to_node = self.nodes[to_node_key]
            to_node_type = to_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if to_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                prop_quantizer.current_location_node_key = to_node_key

        target_ip_node_key = prop_quantizer.current_location_node_key
        target_node = self.nodes[target_ip_node_key]
        target_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = prop_quantizer
        return prop_quantizer

    def add_propagating_quantizer(self, qconf_list: List[QuantizerConfig], ip_node_key: str,
                                  unified_scale: bool = False) -> PropagatingQuantizer:
        prop_quantizer = PropagatingQuantizer(self._get_next_prop_quantizer_id(), qconf_list, ip_node_key,
                                              unified_scale)

        if unified_scale:
            self._unified_scale_group_manager.register_group({prop_quantizer})

        ip_node = self.nodes[ip_node_key]
        ip_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = prop_quantizer
        ip_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(prop_quantizer)

        ip_type = ip_node[QuantizerPropagationStateGraph.INSERTION_POINT_DATA_NODE_ATTR].insertion_type

        if ip_type != InsertionType.OPERATOR_PRE_HOOK:
            # The insertion point key should immediately precede a quantizable op,
            # otherwise it is hard to determine affected node here (although possible)
            raise RuntimeError("Can only add propagating quantizers into pre-hook spots!")

        affected_op_node_key = next(self.successors(ip_node_key))
        affected_op_node = self.nodes[affected_op_node_key]

        affected_op_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(prop_quantizer)

        initial_edge_key = (ip_node_key, affected_op_node_key)
        initial_edge = self.edges[initial_edge_key]
        initial_edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(prop_quantizer)
        prop_quantizer.affected_edges.add(initial_edge_key)
        prop_quantizer.affected_ip_nodes.add(ip_node_key)
        return prop_quantizer

    def clone_propagating_quantizer(self, prop_quantizer: PropagatingQuantizer) -> PropagatingQuantizer:
        cloned_prop_quant = deepcopy(prop_quantizer)
        cloned_prop_quant.id = self._get_next_prop_quantizer_id()
        for edge_tuple in cloned_prop_quant.affected_edges:
            edge = self.edges[edge_tuple]
            edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(cloned_prop_quant)
        for node_key in cloned_prop_quant.affected_ip_nodes:
            node = self.nodes[node_key]
            node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(cloned_prop_quant)

        if cloned_prop_quant.unified_scale:
            gid = self._unified_scale_group_manager.get_group_id_by_propagating_quantizer_id(prop_quantizer.id)
            self._unified_scale_group_manager.add_to_group(gid, cloned_prop_quant)

        return cloned_prop_quant

    def remove_propagating_quantizer(self, prop_quantizer: PropagatingQuantizer):
        for edge_tuple in prop_quantizer.affected_edges:
            edge = self.edges[edge_tuple]
            affecting_quantizers = edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
            affecting_quantizers.remove(prop_quantizer)
        for node_key in prop_quantizer.affected_ip_nodes:
            node = self.nodes[node_key]
            affecting_quantizers = node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
            affecting_quantizers.remove(prop_quantizer)

        node_key = prop_quantizer.current_location_node_key
        self.nodes[node_key][QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = None
        prop_quantizer.affected_ip_nodes.clear()
        prop_quantizer.affected_edges.clear()
        if prop_quantizer.unified_scale:
            gid = self._unified_scale_group_manager.get_group_id_by_propagating_quantizer_id(prop_quantizer.id)
            self._unified_scale_group_manager.remove_from_group(gid, prop_quantizer)

    def propagate_quantizer_via_path(self, prop_quantizer: PropagatingQuantizer, path: List) -> PropagatingQuantizer:
        curr_node_key = prop_quantizer.current_location_node_key
        curr_node = self.nodes[curr_node_key]
        existing_quantizer = curr_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR]
        if existing_quantizer is not None and existing_quantizer.id == prop_quantizer.id:
            curr_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = None
        for edge_tuple in path:
            edge = self.edges[edge_tuple]
            edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(prop_quantizer)
            prop_quantizer.affected_edges.add(edge_tuple)
            prop_quantizer.propagation_path.append(edge_tuple)
            from_node_key = edge_tuple[0]
            from_node = self.nodes[from_node_key]
            from_node_type = from_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if from_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                from_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(prop_quantizer)
                prop_quantizer.affected_ip_nodes.add(from_node_key)
                if self._is_position_accepting(from_node_key):
                    prop_quantizer.last_accepting_location_node_key = from_node_key

        target_ip_node_key = path[-1][0]
        prop_quantizer.current_location_node_key = target_ip_node_key
        target_node = self.nodes[target_ip_node_key]
        target_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] = prop_quantizer
        return prop_quantizer

    def get_quantizable_op_nodes_immediately_dominated_by_node(self, node_key) -> List[str]:
        ret_node_key_list = []

        def recursive_helper(curr_node_key: str, target_node_list: List[str]):
            successors = self.successors(curr_node_key)
            for successor_key in successors:
                successor = self.nodes[successor_key]
                successor_node_type = successor[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
                if successor_node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                    trait = successor[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR]
                    if not trait == QuantizationTrait.QUANTIZATION_AGNOSTIC:
                        target_node_list.append(successor_key)
                        return
                recursive_helper(successor_key, target_node_list)

        recursive_helper(node_key, ret_node_key_list)
        return ret_node_key_list

    def get_paths_to_immediately_dominating_insertion_points(self, insertion_point_node_key: str) -> List[List]:
        """Paths are lists of edges."""
        paths = []

        def recursive_helper(curr_edge, curr_path, all_paths):
            curr_path.append(curr_edge)
            curr_node_key = curr_edge[0]
            curr_node = self.nodes[curr_node_key]
            curr_node_type = curr_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if curr_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                all_paths.append(curr_path)
                return

            for in_edge in self.in_edges(curr_node_key):
                path_copy = deepcopy(curr_path)
                recursive_helper(in_edge, path_copy, all_paths)

        for in_edge in self.in_edges(insertion_point_node_key):
            if self.nodes[in_edge[0]][QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR] == \
                    QuantizerPropagationStateGraphNodeType.AUXILIARY_BARRIER:
                return paths
            recursive_helper(in_edge, [], paths)
        return paths

    def get_visualized_graph(self):
        out_graph = nx.DiGraph()
        unified_scale_group_vs_pq_node_id_dict = {}  # type: Dict[int, List[str]]
        for node_key, node in self.nodes.items():
            node_type = node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                insertion_point_data = node[
                    QuantizerPropagationStateGraph.INSERTION_POINT_DATA_NODE_ATTR]  # type: InsertionPoint
                label = "IP: {}".format(insertion_point_data.insertion_type)
                out_graph.add_node(node_key, label=label, color="red")
                if node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] is not None:
                    prop_quantizer = node[
                        QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR]  # type: PropagatingQuantizer
                    quant_node_key = "Quantizer #{}".format(prop_quantizer.id)
                    if prop_quantizer.potential_quant_configs:
                        quant_configs_str_list = [str(conf) for conf in prop_quantizer.potential_quant_configs]
                    else:
                        quant_configs_str_list = ["!!! NONE !!!]"]
                    sub_label = '[' + ',\n'.join(quant_configs_str_list) + ']'
                    quant_node_label = quant_node_key + '\n' + "T: {}\n".format(sub_label)
                    out_graph.add_node(quant_node_key,
                                       color="blue", label=quant_node_label)
                    out_graph.add_edge(quant_node_key, node_key,
                                       style="dashed")
                    if prop_quantizer.unified_scale:
                        gid = self._unified_scale_group_manager.get_group_id_by_propagating_quantizer_id(
                            prop_quantizer.id)
                        if gid in unified_scale_group_vs_pq_node_id_dict:
                            unified_scale_group_vs_pq_node_id_dict[gid].append(quant_node_key)
                        else:
                            unified_scale_group_vs_pq_node_id_dict[gid] = [quant_node_key]

            elif node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                out_graph.add_node(node_key)
            elif node_type == QuantizerPropagationStateGraphNodeType.AUXILIARY_BARRIER:
                out_graph.add_node(node_key, color='green', label=node['label'])
            else:
                raise RuntimeError("Invalid QuantizerPropagationStateGraph node!")
        for u, v in self.edges:
            edge = self.edges[u, v]
            attrs = {}
            affecting_quantizers = edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
            if affecting_quantizers:
                label = ", ".join([str(pq.id) for pq in affecting_quantizers])
                attrs = {"color": "blue", "label": label}
            out_graph.add_edge(u, v, **attrs)

        for gid, group_pq_node_keys in unified_scale_group_vs_pq_node_id_dict.items():
            if len(group_pq_node_keys) < 2:
                continue
            curr_elt_iter = iter(group_pq_node_keys)
            next_elt_iter = iter(group_pq_node_keys)
            _ = next(next_elt_iter)  # in order not to use more_itertools.consume
            done = False
            while not done:
                curr_pq_node_key = next(curr_elt_iter)
                try:
                    next_pq_node_key = next(next_elt_iter)
                except StopIteration:
                    done = True
                    next_pq_node_key = group_pq_node_keys[0]  # back to the first elt
                out_graph.add_edge(curr_pq_node_key, next_pq_node_key, arrowhead="none",
                                   style="dotted",
                                   label="Unified group {}".format(gid))

        return out_graph

    def traverse_graph(self, curr_node_key: str,
                       traverse_function: Callable[[str, Any], Tuple[bool, Any]],
                       output: Any,
                       traverse_forward: bool = True) -> Any:
        return self._traverse_graph_recursive_helper(curr_node_key, traverse_function, output, traverse_forward)

    def _traverse_graph_recursive_helper(self, curr_node_key: str,
                                         traverse_function: Callable[[str, Any], Tuple[bool, Any]],
                                         output: Any, traverse_forward: bool):
        is_finished, output = traverse_function(curr_node_key, output)
        node_keys_holder = self.succ if traverse_forward else self.pred
        if not is_finished:
            for node_key in node_keys_holder[curr_node_key]:
                self._traverse_graph_recursive_helper(node_key, traverse_function, output, traverse_forward)
        return output

    def _get_next_prop_quantizer_id(self):
        self._created_prop_quantizer_counter += 1
        return self._created_prop_quantizer_counter

    def _is_position_accepting(self, ip_node_key: str):
        node = self.nodes[ip_node_key]
        insertion_type = node[QuantizerPropagationStateGraph.INSERTION_POINT_DATA_NODE_ATTR].insertion_type
        if insertion_type == InsertionType.OPERATOR_POST_HOOK:
            return True
        return False

    def get_unified_scale_group_id_by_propagating_quantizer_id(self, pqid: int) -> int:
        return self._unified_scale_group_manager.get_group_id_by_propagating_quantizer_id(pqid)

    def get_input_quantizer_ids(self) -> Dict[InputAgnosticOperationExecutionContext, List[int]]:
        retval = {}  # type: Dict[InputAgnosticOperationExecutionContext, List[int]]

        def recursive_helper(curr_node_key: str, curr_input_quantizer_ids_list: List[int]):
            curr_node = self.nodes[curr_node_key]
            curr_node_type = curr_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]

            if curr_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                pq = curr_node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR]
                if pq is not None:
                    curr_input_quantizer_ids_list.append(pq.id)
                    return
            elif curr_node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                trait = curr_node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR]
                if not trait == QuantizationTrait.QUANTIZATION_AGNOSTIC:
                    return
            elif curr_node_type == QuantizerPropagationStateGraphNodeType.AUXILIARY_BARRIER:
                return

            for successor_key in self.successors(curr_node_key):
                recursive_helper(successor_key, curr_input_quantizer_ids_list)

        for input_node_key, input_node_context in self._input_node_keys_vs_contexts.items():
            current_input_quantizer_ids = []
            recursive_helper(input_node_key, current_input_quantizer_ids)
            retval[input_node_context] = current_input_quantizer_ids

        return retval


class QuantizersBetweenQuantizableLayers:
    """ Contains locations of quantizers between inputs quantizable layers: input agnostic operation execution context
    for activations and scope - for quantized modules """

    def __init__(self):
        self.activation_quantizer_ctxs = set()  # type: Set[InputAgnosticOperationExecutionContext]
        self.quantized_module_scopes = set()  # type: Set['Scope']

    def add_activation_quantizer_ctx(self, iap_ctx: InputAgnosticOperationExecutionContext):
        self.activation_quantizer_ctxs.add(iap_ctx)

    def add_quantized_module_scope(self, scope: 'Scope'):
        self.quantized_module_scopes.add(scope)

    def __bool__(self) -> bool:
        return bool(self.activation_quantizer_ctxs) and bool(self.quantized_module_scopes)


class QuantizerPropagationSolver:
    """Analyzes a fresh QuantizerPropagationStateGraph object according to HW
       configuration supplied in the initializer and produces the list of insertion
       commands that correspond to the final state of the quantizer propagation graph
       when the model has the most contol flow graph edges quantized according to HW
       capabilities."""

    DEFAULT_QUANTIZATION_TYPES = [QuantizerConfig(
        bits=8,
        mode=QuantizationMode.SYMMETRIC,
        signedness_to_force=None,
        per_channel=False)]

    def __init__(self, ignored_scopes=None, hw_config: HWConfig = None,
                 debug_interface: 'QuantizationDebugInterface' = None,
                 propagation_strategy: PropagationStrategy = PropagationStrategy.AGGRESSIVE,
                 default_qconfig_list: List[QuantizerConfig] = None,
                 input_infos: List[ModelInputInfo] = None):
        self.quantizers_between_quantizable_layers_per_key = {}  # type: Dict[str, QuantizersBetweenQuantizableLayers]
        self.default_qlobal_qconfig_list = default_qconfig_list
        self._hw_config = hw_config  # type: HWConfig
        self._debug_interface = debug_interface
        self._propagation_strategy = propagation_strategy  # TODO: determine from config
        self._operator_quantization_trait_map = self.get_operator_quantization_traits_map()
        self._operator_allowed_qconfigs_map = self._get_operator_qconfigs_map()
        self._input_infos = input_infos

        if self._hw_config is not None:
            self._unified_scales_operation_set = self._hw_config.get_operations_with_unified_scales()
        else:
            self._unified_scales_operation_set = {}

        # Will handle the "wildcard" quantization situation for the time being
        if default_qconfig_list is not None:
            for op_meta, qconf_list in self._operator_allowed_qconfigs_map.items():
                trait = self._operator_quantization_trait_map[op_meta]
                if trait == QuantizationTrait.INPUTS_QUANTIZABLE:
                    # !!! FIXME !!! Ensure that INPUTS_QUANTIZABLE ops always have non-None qconf_list
                    if HWConfig.is_qconf_list_corresponding_to_unspecified_op(qconf_list):
                        self._operator_allowed_qconfigs_map[op_meta] = default_qconfig_list
        self._active_propagating_quantizers_queue = deque()
        self._finished_propagating_quantizers = []  # type: List[PropagatingQuantizer]
        self._ignored_scopes = ignored_scopes

        self._potential_quantizers = {}

    def run_on_ip_graph(self, ip_graph: InsertionPointGraph) -> Dict[QuantizerInsertionInfo,
                                                                     Optional[List[QuantizerConfig]]]:
        """ The main function to be used on an InsertionPointGraph to produce
            the list of insertion commands and configs corresponding to the final quantized
            graph state."""
        quant_prop_graph = QuantizerPropagationStateGraph(ip_graph, self._ignored_scopes)
        quant_prop_graph = self.set_allowed_quantization_types_for_operator_nodes(quant_prop_graph)
        quant_prop_graph = self.setup_initial_quantizers(quant_prop_graph)
        iteration_counter = 0
        while self._active_propagating_quantizers_queue:
            prop_quantizer = self._active_propagating_quantizers_queue.pop()
            if self._debug_interface is not None:
                self._debug_interface.visualize_quantizer_propagation(self, quant_prop_graph, str(iteration_counter))
            quant_prop_graph = self.propagation_step(prop_quantizer, quant_prop_graph)
            iteration_counter += 1

        if self._input_infos is not None:
            self._filter_integer_input_quantizers(quant_prop_graph)

        if self._debug_interface is not None:
            self._debug_interface.visualize_quantizer_propagation(self, quant_prop_graph, "final")

        retval = {}

        non_unified_final_prop_quantizers = set()  # type: Set[PropagatingQuantizer]
        unified_final_prop_quantizers = {}  # type: Dict[int, Set[PropagatingQuantizer]]

        self.quantizers_between_quantizable_layers_per_key = \
            self._get_quantizers_between_quantizable_layers_per_node_key(
                quant_prop_graph, self._finished_propagating_quantizers)

        for finished_prop_quantizer in self._finished_propagating_quantizers:
            if finished_prop_quantizer.unified_scale:
                # Handle unified scale quantizers separately since they require special InsertionInfo construction
                gid = quant_prop_graph.get_unified_scale_group_id_by_propagating_quantizer_id(
                    finished_prop_quantizer.id)
                if gid not in unified_final_prop_quantizers:
                    unified_final_prop_quantizers[gid] = set()
                unified_final_prop_quantizers[gid].add(finished_prop_quantizer)
            else:
                non_unified_final_prop_quantizers.add(finished_prop_quantizer)

        for non_unified_prop_quant in non_unified_final_prop_quantizers:
            insertion_info = self._get_insertion_info_for_propagating_quantizer(non_unified_prop_quant,
                                                                                quant_prop_graph)
            retval[insertion_info] = non_unified_prop_quant.potential_quant_configs

        for unified_prop_quantizer_set in unified_final_prop_quantizers.values():
            insertion_infos_vs_qconfigs = {}  # type: Dict[QuantizerInsertionInfo, List[QuantizerConfig]]
            for prop_quant in unified_prop_quantizer_set:
                insertion_info = self._get_insertion_info_for_propagating_quantizer(prop_quant, quant_prop_graph)
                insertion_infos_vs_qconfigs[insertion_info] = prop_quant.potential_quant_configs

            # The primary insertion point (to be associated with the actual quantizer module, not just hooks to it)
            # will be determined based on the string representation of said insertion point, to avoid random selection
            all_insertion_infos = sorted(list(insertion_infos_vs_qconfigs.keys()),
                                         key=lambda x: str(x.op_exec_context.input_agnostic))
            primary_insertion_info = all_insertion_infos[0]
            linked_insertion_infos = all_insertion_infos[1:]
            primary_insertion_info.linked_op_exec_contexts = [x.op_exec_context for x in linked_insertion_infos]
            retval[primary_insertion_info] = insertion_infos_vs_qconfigs[primary_insertion_info]

        return retval

    def _get_insertion_info_for_propagating_quantizer(self, prop_quant: PropagatingQuantizer,
                                                      quant_prop_graph: QuantizerPropagationStateGraph) -> \
            QuantizerInsertionInfo:
        final_node_key = prop_quant.current_location_node_key
        final_node = quant_prop_graph.nodes[final_node_key]
        insertion_point = final_node[
            QuantizerPropagationStateGraph.INSERTION_POINT_DATA_NODE_ATTR]  # type: InsertionPoint
        # TODO: fix this, rethink InsertionInfo here and elsewhere
        insertion_info = QuantizerInsertionInfo(OperationExecutionContext(
            operator_name=insertion_point.ia_op_exec_context.operator_name,
            scope_in_model=insertion_point.ia_op_exec_context.scope_in_model,
            call_order=insertion_point.ia_op_exec_context.call_order,
            tensor_metas=[None],
        ), quantizers_between_quantizable_layers=self.quantizers_between_quantizable_layers_per_key[final_node_key])
        return insertion_info

    def propagation_step(self, curr_prop_quantizer: PropagatingQuantizer,
                         quant_prop_graph: QuantizerPropagationStateGraph) -> QuantizerPropagationStateGraph:
        """Returns an updated curr_prop_quantizer state if the quantizer is not
           yet in its final (accepting) position, and None if the quantizer is in its
           final location.  The location before and after the step should correspond to
           some insertion point."""
        # TODO: full-fledged discrete finite automata approach? Switch to traversing a graph
        # consisting of insertion points only, with reversed edges holding associated operator data?
        curr_node_key = curr_prop_quantizer.current_location_node_key
        curr_node = quant_prop_graph.nodes[curr_prop_quantizer.current_location_node_key]
        curr_node_type = curr_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
        assert curr_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT

        # Assumption: paths are at most 2 edges - either one edge to neighbouring insertion point
        # or one edge to operation and next edge to its own neighbouring insertion point.

        paths = quant_prop_graph.get_paths_to_immediately_dominating_insertion_points(curr_node_key)
        if not paths:
            prop_quantizer = quant_prop_graph.backtrack_propagation_until_accepting_location(curr_prop_quantizer)
            if prop_quantizer is not None:
                self._finished_propagating_quantizers.append(prop_quantizer)
            return quant_prop_graph

        surviving_prop_quantizers = []

        prop_quantizers_to_process = [curr_prop_quantizer]
        for _ in range(1, len(paths)):
            additional_prop_quantizer = quant_prop_graph.clone_propagating_quantizer(curr_prop_quantizer)
            prop_quantizers_to_process.append(additional_prop_quantizer)

        pqs_and_paths = zip(paths, prop_quantizers_to_process)
        for path, prop_quantizer in pqs_and_paths:
            status = self.check_transition_via_path(prop_quantizer, path, quant_prop_graph)
            if status == TransitionStatus.SHOULD_NOT_TRANSITION:
                prop_quantizer = quant_prop_graph.backtrack_propagation_until_accepting_location(prop_quantizer)
                if prop_quantizer is not None:
                    self._finished_propagating_quantizers.append(prop_quantizer)
            elif status == TransitionStatus.SHOULD_TRANSITION:
                prop_quantizer = quant_prop_graph.propagate_quantizer_via_path(prop_quantizer, path)
                surviving_prop_quantizers.append(prop_quantizer)
            elif status == TransitionStatus.SHOULD_MERGE:
                # The surviving quantizer will have its "affected edges" set extended
                # by the corresponding set of the merged quantizer. The assumption
                # here is that the surviving quantizer should never have to cross
                # such a "merge point" while backtracking to an accepting location.

                quant_prop_graph.merge_quantizer_into_path(prop_quantizer, path)

        for prop_quantizer in surviving_prop_quantizers:
            self._active_propagating_quantizers_queue.appendleft(prop_quantizer)
        return quant_prop_graph

    def get_allowed_quantizer_configs_for_operator(self, quant_det_id: OperatorMetatype) -> List[QuantizerConfig]:
        return self._operator_allowed_qconfigs_map[quant_det_id]

    def set_allowed_quantization_types_for_operator_nodes(self, quant_prop_graph: QuantizerPropagationStateGraph):
        for node_key, node in quant_prop_graph.nodes.items():
            node_type = node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                quant_det_id = node[QuantizerPropagationStateGraph.OPERATOR_METATYPE_NODE_ATTR]
                if quant_det_id is None:
                    warnings.warn("Unknown metatype for operator node: {}".format(node_key))
                    trait = QuantizationTrait.QUANTIZATION_AGNOSTIC
                else:
                    trait = self._operator_quantization_trait_map[quant_det_id]
                node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR] = trait
                if trait == QuantizationTrait.INPUTS_QUANTIZABLE:
                    node[QuantizerPropagationStateGraph.ALLOWED_INPUT_QUANTIZATION_TYPES_NODE_ATTR] = \
                        self.get_allowed_quantizer_configs_for_operator(quant_det_id)
        return quant_prop_graph

    def get_operator_quantization_traits_map(self) -> Dict[OperatorMetatype, QuantizationTrait]:
        # TODO: ensure that there are no name collisions between ops in different torch subpackages with the same name
        retval = {}
        if self._hw_config is None:
            for op_meta in OPERATOR_METATYPES.registry_dict.values():
                retval[op_meta] = QuantizationTrait.QUANTIZATION_AGNOSTIC  # Default value
            for trait, meta_list in DEFAULT_QUANT_TRAIT_TO_OP_DICT.items():
                for op_meta in meta_list:  # type: OperatorMetatype
                    retval[op_meta] = trait
        else:
            op_meta_vs_qconfs_map = self._hw_config.get_metatype_vs_quantizer_configs_map()
            for op_meta, qconf_list in op_meta_vs_qconfs_map.items():
                if HWConfig.is_qconf_list_corresponding_to_unspecified_op(qconf_list):
                    trait = self._get_trait_for_op_meta_not_specified_in_hw_config(op_meta)
                elif HWConfig.is_wildcard_quantization(qconf_list):
                    for default_trait, meta_list in DEFAULT_QUANT_TRAIT_TO_OP_DICT.items():
                        if op_meta in meta_list:
                            trait = default_trait
                            break
                    else:
                        trait = QuantizationTrait.QUANTIZATION_AGNOSTIC
                else:
                    trait = QuantizationTrait.INPUTS_QUANTIZABLE
                retval[op_meta] = trait
        return retval

    @staticmethod
    def _get_trait_for_op_meta_not_specified_in_hw_config(op_meta: OperatorMetatype) -> QuantizationTrait:
        if not op_meta.hw_config_names:
            # The metatype might not have an associated name in the config
            # namespace (yet) - use default trait
            for default_trait, meta_list in DEFAULT_QUANT_TRAIT_TO_OP_DICT.items():
                if op_meta in meta_list:
                    trait = default_trait
                    break
            else:
                trait = QuantizationTrait.QUANTIZATION_AGNOSTIC
                # TODO: think of switching to this?
                # raise RuntimeError("Operation metatype {} encountered, but it has no default "
                #                    "quantization trait and the HW config entry is not given for it - "
                #                    "cannot determine how to quantize it!".format(op_meta))
        else:
            # There IS a valid HW config name for the metatype, but it is deliberately not specified
            # in the config, which means that it should execute in FP32
            trait = QuantizationTrait.NON_QUANTIZABLE

        return trait

    def _get_operator_qconfigs_map(self) -> Dict[OperatorMetatype, List[QuantizerConfig]]:
        # TODO: ensure that there are no name collisions between ops in different torch subpackages with the same name
        retval = {}
        if self._hw_config is None:
            for op_meta in OPERATOR_METATYPES.registry_dict.values():
                retval[op_meta] = []  # Default value, corresponds to wildcard quantization
            for trait, meta_list in DEFAULT_QUANT_TRAIT_TO_OP_DICT.items():
                if trait == QuantizationTrait.INPUTS_QUANTIZABLE:
                    for op_meta in meta_list:  # type: OperatorMetatype
                        if self.default_qlobal_qconfig_list is not None:
                            retval[op_meta] = deepcopy(self.default_qlobal_qconfig_list)
                        else:
                            retval[op_meta] = deepcopy(self.DEFAULT_QUANTIZATION_TYPES)
                elif trait == QuantizationTrait.NON_QUANTIZABLE:
                    for op_meta in meta_list:  # type: OperatorMetatype
                        retval[op_meta] = None
        else:
            retval = self._hw_config.get_metatype_vs_quantizer_configs_map()
        return retval

    def debug_visualize(self, quant_prop_graph: QuantizerPropagationStateGraph, dump_path: str):
        out_graph = quant_prop_graph.get_visualized_graph()
        active_ids_str = ", ".join([str(pq.id) for pq in self._active_propagating_quantizers_queue])
        finished_ids_str = ", ".join([str(pq.id) for pq in self._finished_propagating_quantizers])
        out_graph.graph['graph'] = {"label": "Propagating quantizers: {}\n" \
                                             "Finished quantizers: {}".format(active_ids_str, finished_ids_str),
                                    "labelloc": "t"}
        nx.drawing.nx_pydot.write_dot(out_graph, dump_path)

    def setup_initial_quantizers(self,
                                 quant_prop_graph: QuantizerPropagationStateGraph) -> QuantizerPropagationStateGraph:
        """Determines the initial subset of the nodes that must be quantized
           and corresponding allowed quantization configs (possibly multiple) for each
           quantizer."""
        for node_key, node in quant_prop_graph.nodes.items():
            node_type = node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                if node_key in quant_prop_graph.ignored_node_keys:
                    continue

                preds = list(quant_prop_graph.predecessors(node_key))
                if not preds:
                    continue  # TODO: remove this once module insertion points are included in the IP graph
                # Should be immediately preceded by an insertion point.
                pred_ip_key = preds[0]
                pred_node = quant_prop_graph.nodes[pred_ip_key]
                pred_node_type = pred_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
                assert pred_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT, \
                    "Invalid insertion point graph supplied for quantizer propagation!"

                if node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR] in [
                        QuantizationTrait.NON_QUANTIZABLE,
                        QuantizationTrait.QUANTIZATION_AGNOSTIC]:
                    continue

                quant_det_id = node[QuantizerPropagationStateGraph.OPERATOR_METATYPE_NODE_ATTR]
                qconf_list = self.get_allowed_quantizer_configs_for_operator(quant_det_id)

                # No need to place quantizers for FP32-forced ops, naturally
                assert qconf_list is not None

                is_unified_scale = quant_det_id in self._unified_scales_operation_set
                if is_unified_scale:
                    # Filtering out the per-channel cases in the unified scale scenario.
                    # In order to support unified per-channel scales, we will need to handle a situation
                    # when a per-channel shared (unified) quantizer on one branch passes a shape-chaning
                    # operation such as `view` or `transpose`, and the linked unified quantizer does not do
                    # so due to one or the other reason; the per-channel scale shapes to be applied therefore
                    # will be different.
                    # TODO: What possibly needs to be done in this direction:
                    # 1. keep ForwardTraceOnly ops in graph after all, to be able to track shape changes
                    # 2. transpose input tensors to the quantization modules on the fly to accomodate scale,
                    #    or vice versa, transpose scale to accomodate shape; need to handle exporting as well
                    per_tensor_qconf_list = list(filter(lambda x: x.per_channel is False, qconf_list))
                    op_meta_name = quant_det_id.__class__.__name__
                    if len(per_tensor_qconf_list) != len(qconf_list):
                        if not per_tensor_qconf_list:
                            raise RuntimeError(
                                "Unified scales currently do not support per-channel configuration - dropping"
                                "per-channel configuration options for {} resulted in no valid quantization "
                                "configs!".format(op_meta_name))
                        nncf_logger.warning(
                            "Unified scales currently do not support per-channel configuration - dropping"
                            "per-channel configuration options for {}".format(op_meta_name))
                        qconf_list = per_tensor_qconf_list

                prop_quantizer = quant_prop_graph.add_propagating_quantizer(qconf_list, pred_ip_key, is_unified_scale)
                self._active_propagating_quantizers_queue.appendleft(prop_quantizer)

        return quant_prop_graph

    # pylint:disable=too-many-return-statements
    def check_branching_transition(self, quant_prop_graph: QuantizerPropagationStateGraph,
                                   prop_quantizer: PropagatingQuantizer,
                                   branching_node_key: str) -> Optional[TransitionStatus]:
        """If a propagating quantizer advances through a node that branches
           downwards, the branches neighbouring to the one that the propagating quantizer
           had just propagated from will have the precision of the quantizer imposed upon
           them.  This is not always desirable - we might want to keep some branches in
           higher precision than the others. For this reason, this function checks whether
           the quantizer may safely advance through a branching node based on the possible
           configs of the quantizers it might affect by doing so."""
        dom_op_node_keys = quant_prop_graph.get_quantizable_op_nodes_immediately_dominated_by_node(
            branching_node_key)
        primary_possible_qconfigs = prop_quantizer.potential_quant_configs
        secondary_possible_qconfigs_dict = {}
        for op_node_key in dom_op_node_keys:
            op_node = quant_prop_graph.nodes[op_node_key]
            affecting_prop_quantizers = op_node[
                QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
            if not affecting_prop_quantizers:
                # The branch op is forced to be FP32 - should not proceed through the branch node.
                return TransitionStatus.SHOULD_NOT_TRANSITION
            secondary_possible_qconfigs = affecting_prop_quantizers[0].potential_quant_configs
            secondary_possible_qconfigs_dict[op_node_key] = secondary_possible_qconfigs
        primary_merged_qconfigs, \
        secondary_merged_qconfigs_dict = self.get_merged_qconfigs(primary_possible_qconfigs,
                                                                  secondary_possible_qconfigs_dict)
        if not primary_merged_qconfigs:
            # This quantizer's precision does not encompass the precisions of quantizers
            # propagating through downward branches.
            return TransitionStatus.SHOULD_NOT_TRANSITION

        if self._propagation_strategy == PropagationStrategy.CONSERVATIVE:
            for op_node_key, secondary_merged_qconfigs_list in secondary_merged_qconfigs_dict.items():
                if len(secondary_possible_qconfigs_dict[op_node_key]) != len(secondary_merged_qconfigs_list):
                    return TransitionStatus.SHOULD_NOT_TRANSITION

        return None

    def check_transition_via_path(self, prop_quantizer: PropagatingQuantizer, path: List,
                                  quant_prop_graph: QuantizerPropagationStateGraph) -> TransitionStatus:
        """Determines which action should be taken regarding the
           prop_quantizer's propagation via path, which may be one of many possible
           propagation paths."""
        for from_node_key, to_node_key in path:
            from_node = quant_prop_graph.nodes[from_node_key]
            if len(list(quant_prop_graph.successors(from_node_key))) > 1:
                # If a quantizer simply passes up through a downward-branching node, it may spoil the
                # precision for operations on neighbouring branches. Consider a 4-bit quantizer rising
                # through a branch node and an 8-bit quantizer arriving at the same node later. Therefore,
                # prior to allowing the quantizer to pass through a branching node we need to ensure that
                # the precision of the quantizer is a superset of precisions of the first non-quantization agnostic
                # operations on each branch.
                status = self.check_branching_transition(quant_prop_graph,
                                                         prop_quantizer,
                                                         from_node_key)
                if status is not None:
                    return status

            # Check if current edge to traverse is affected by any of the quantizers
            from_node_type = from_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if from_node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                trait = from_node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR]
                if trait in [QuantizationTrait.NON_QUANTIZABLE,
                             QuantizationTrait.INPUTS_QUANTIZABLE]:
                    return TransitionStatus.SHOULD_NOT_TRANSITION
            edge = quant_prop_graph.edges[from_node_key, to_node_key]
            potential_quantizers = edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
            if potential_quantizers:
                # Assuming that multiple affecting quantizers should all have the same quantization config
                # by construction
                curr_pq_configs = prop_quantizer.potential_quant_configs
                target_pq_configs = potential_quantizers[0].potential_quant_configs
                if curr_pq_configs == target_pq_configs or \
                        HWConfig.is_wildcard_quantization(curr_pq_configs) or \
                        HWConfig.is_wildcard_quantization(target_pq_configs):
                    return TransitionStatus.SHOULD_MERGE
                return TransitionStatus.SHOULD_NOT_TRANSITION

            # Check if the target node is affected by any of the quantizers
            from_node_type = from_node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            if from_node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                potential_quantizers = from_node[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR]
                if potential_quantizers:
                    # Affecting quantizers should have the same configs by construction, so we only
                    # check the first
                    curr_pq_configs = prop_quantizer.potential_quant_configs
                    target_pq_configs = potential_quantizers[0].potential_quant_configs
                    if curr_pq_configs == target_pq_configs or \
                            HWConfig.is_wildcard_quantization(curr_pq_configs) or \
                            HWConfig.is_wildcard_quantization(target_pq_configs):
                        return TransitionStatus.SHOULD_MERGE

                    # Did not merge - the edge will remain untraversed, but the quantizers at the next node will
                    # still be affecting it
                    for pq in potential_quantizers:
                        pq.affected_edges.add((from_node_key, to_node_key))
                        edge[QuantizerPropagationStateGraph.AFFECTING_PROPAGATING_QUANTIZERS_ATTR].append(pq)

                    return TransitionStatus.SHOULD_NOT_TRANSITION
        return TransitionStatus.SHOULD_TRANSITION

    def get_merged_qconfigs(self, primary_potential_qconfigs_list: List[QuantizerConfig],
                            secondary_potential_qconfigs_dict: Dict[str, List[QuantizerConfig]]) -> Tuple[
                                List[QuantizerConfig], Dict[str, QuantizerConfig]]:
        """Returns potential qconfigs lists for 'primary' and 'secondary' quantizers
        that are compatible with each other. Compatibility is decided in terms of
        primary quantizer having configs which all have higher precision than all the
        secondary potential quantizer configs."""
        final_primary_merged_qconfigs_list = deepcopy(primary_potential_qconfigs_list)
        curr_secondary_merged_qconfigs_dict = deepcopy(secondary_potential_qconfigs_dict)
        # TODO: implement variant solutions, i.e. for each set of resultant merged
        # primary potential qconfig lists we have, in general, different merged secondary potential
        # config lists. Currently greedy approach is used.
        for m_qconfig in primary_potential_qconfigs_list:
            should_persist_secondary_merged_qconfigs_dict = True
            candidate_secondary_merged_qconfigs_dict = deepcopy(curr_secondary_merged_qconfigs_dict)
            for node_key, s_qconfig_list in curr_secondary_merged_qconfigs_dict.items():
                for s_qconfig in s_qconfig_list:
                    if m_qconfig < s_qconfig and s_qconfig in candidate_secondary_merged_qconfigs_dict[node_key]:
                        candidate_secondary_merged_qconfigs_dict[node_key].remove(s_qconfig)
            for _, s_qconfig_list in candidate_secondary_merged_qconfigs_dict.items():
                if not s_qconfig_list:
                    # No options left for secondary configs on one of the branches to accomodate the primary
                    # config - this primary config cannot be used to be merged into.
                    final_primary_merged_qconfigs_list.remove(m_qconfig)
                    should_persist_secondary_merged_qconfigs_dict = False
                    break
            if should_persist_secondary_merged_qconfigs_dict:
                curr_secondary_merged_qconfigs_dict = candidate_secondary_merged_qconfigs_dict
        if not final_primary_merged_qconfigs_list:
            return [], {}
        return final_primary_merged_qconfigs_list, curr_secondary_merged_qconfigs_dict

    def get_finished_propagating_quantizers(self):
        return self._finished_propagating_quantizers

    def get_active_propagating_quantizers_queue(self):
        return self._active_propagating_quantizers_queue

    def _filter_integer_input_quantizers(self, quant_prop_graph: QuantizerPropagationStateGraph):
        input_node_vs_qid_dict = quant_prop_graph.get_input_quantizer_ids()
        integer_input_quantizer_ids = set()

        for input_node_context, input_quantizer_ids in input_node_vs_qid_dict.items():
            assert input_node_context.operator_name == MODEL_INPUT_OP_NAME
            input_id = input_node_context.call_order
            if self._input_infos[input_id].is_integer_input():
                integer_input_quantizer_ids.update(set(input_quantizer_ids))

        filtered_finished_pqs = list(filter(lambda pq: pq.id not in integer_input_quantizer_ids,
                                            self._finished_propagating_quantizers))
        self._finished_propagating_quantizers = filtered_finished_pqs

    @staticmethod
    def _get_quantizers_between_quantizable_layers_per_node_key(
            quant_prop_graph: QuantizerPropagationStateGraph,
            finished_propagating_quantizers: List[PropagatingQuantizer]):
        visited = {node_key: False for node_key in quant_prop_graph.nodes()}
        quantizers_between_quantizable_layers_per_node_key = {}  # type: Dict[str, QuantizersBetweenQuantizableLayers]

        def traverse_function_up(node_key: str,
                                 output: QuantizersBetweenQuantizableLayers) -> Tuple[bool, Any]:
            if visited[node_key]:
                return True, output
            visited[node_key] = True

            is_finished = False
            node = quant_prop_graph.nodes[node_key]
            node_type = node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]

            if node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                insertion_point_data = node[
                    QuantizerPropagationStateGraph.INSERTION_POINT_DATA_NODE_ATTR]  # type: InsertionPoint
                if node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] is not None:
                    output.add_activation_quantizer_ctx(insertion_point_data.ia_op_exec_context)
                    quantizers_between_quantizable_layers_per_node_key[node_key] = output
                    is_finished = True
                else:
                    for sub_node_key in quant_prop_graph.succ[node_key]:
                        output = quant_prop_graph.traverse_graph(sub_node_key, traverse_function_down, output,
                                                                 traverse_forward=True)
            elif node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                if node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR] \
                    == QuantizationTrait.INPUTS_QUANTIZABLE:
                    raise RuntimeError('Should not reach quantizable operator on backward traverse from quantizer!')
            else:
                # reached barrier for nodes in ignored_scopes, no need to go further - this nodes shouldn't be quantized
                is_finished = True
            return is_finished, output

        def traverse_function_down(node_key: str, output: QuantizersBetweenQuantizableLayers) -> Tuple[bool, Any]:
            if visited[node_key]:
                return True, output
            visited[node_key] = True

            node = quant_prop_graph.nodes[node_key]
            node_type = node[QuantizerPropagationStateGraph.NODE_TYPE_NODE_ATTR]
            is_finished = False
            if node_type == QuantizerPropagationStateGraphNodeType.INSERTION_POINT:
                insertion_point_data = node[
                    QuantizerPropagationStateGraph.INSERTION_POINT_DATA_NODE_ATTR]  # type: InsertionPoint
                if node[QuantizerPropagationStateGraph.PROPAGATING_QUANTIZER_NODE_ATTR] is not None:
                    output.add_activation_quantizer_ctx(insertion_point_data.ia_op_exec_context)
                    quantizers_between_quantizable_layers_per_node_key[node_key] = output
                else:
                    for sub_node_key in quant_prop_graph.pred[node_key]:
                        output = quant_prop_graph.traverse_graph(sub_node_key, traverse_function_up, output,
                                                                 traverse_forward=False)
            elif node_type == QuantizerPropagationStateGraphNodeType.OPERATOR:
                if node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR] \
                    == QuantizationTrait.INPUTS_QUANTIZABLE:
                    output.add_quantized_module_scope(node[QuantizerPropagationStateGraph.OPERATOR_SCOPE])
                    is_finished = True
                elif node[QuantizerPropagationStateGraph.QUANTIZATION_TRAIT_NODE_ATTR] \
                    == QuantizationTrait.NON_QUANTIZABLE:
                    raise RuntimeError('Should not reach non-quantizable operator on forward traverse from quantizer!')
            else:
                # reached barrier for nodes in ignored_scopes, no need to go further - this nodes shouldn't be quantized
                is_finished = True
            return is_finished, output

        for finished_prop_quantizer in finished_propagating_quantizers:
            node_key = finished_prop_quantizer.current_location_node_key
            quantizers_between_quantizable_layers = QuantizersBetweenQuantizableLayers()
            quant_prop_graph.traverse_graph(node_key, traverse_function_down, quantizers_between_quantizable_layers,
                                            traverse_forward=True)

        return quantizers_between_quantizable_layers_per_node_key
